from __future__ import absolute_import, division, unicode_literals

import logging
import os
from builtins import map
from datetime import datetime, timedelta

from past.utils import old_div

from .config import *
from .constants import *
from .database import Database
from .encoder import MetaData
from .method import Method
from .utilities import download_data

# load the library-wide logger
logger = logging.getLogger('atm')


def create_dataset(db, run_config, aws_config=None):
    """
    Create a dataset and add it to the ModelHub database.

    db: initialized Database object
    run_config: RunConfig object describing the dataset to create
    aws_config: optional. AWS credentials for downloading data from S3.
    """
    # download data to the local filesystem to extract metadata
    train_local, test_local = download_data(run_config.train_path,
                                            run_config.test_path,
                                            aws_config)

    # create the name of the dataset from the path to the data
    name = os.path.basename(train_local)
    name = name.replace("_train.csv", "").replace(".csv", "")

    # process the data into the form ATM needs and save it to disk
    meta = MetaData(run_config.class_column, train_local, test_local)

    # enter dataset into database
    dataset = db.create_dataset(name=name,
                                description=run_config.data_description,
                                train_path=run_config.train_path,
                                test_path=run_config.test_path,
                                class_column=run_config.class_column,
                                n_examples=meta.n_examples,
                                k_classes=meta.k_classes,
                                d_features=meta.d_features,
                                majority=meta.majority,
                                size_kb=old_div(meta.size, 1000))
    return dataset


def create_datarun(db, dataset, run_config):
    """
    Given a config, creates a set of dataruns for the config and enters them into
    the database. Returns the ID of the created datarun.

    db: initialized Database object
    dataset: Dataset SQLAlchemy ORM object
    run_config: RunConfig object describing the datarun to create
    """
    # describe the datarun by its tuner and selector
    run_description = '__'.join([run_config.tuner, run_config.selector])

    # set the deadline, if applicable
    deadline = run_config.deadline
    if deadline:
        deadline = datetime.strptime(deadline, TIME_FMT)
        # this overrides the otherwise configured budget_type
        # TODO: why not walltime and classifiers budget simultaneously?
        run_config.budget_type = 'walltime'
    elif run_config.budget_type == 'walltime':
        deadline = datetime.now() + timedelta(minutes=budget)

    target = run_config.score_target + '_judgment_metric'
    datarun = db.create_datarun(dataset_id=dataset.id,
                                description=run_description,
                                tuner=run_config.tuner,
                                selector=run_config.selector,
                                gridding=run_config.gridding,
                                priority=run_config.priority,
                                budget_type=run_config.budget_type,
                                budget=run_config.budget,
                                deadline=deadline,
                                metric=run_config.metric,
                                score_target=target,
                                k_window=run_config.k_window,
                                r_minimum=run_config.r_minimum)
    return datarun


def enter_data(sql_config, run_config, aws_config=None,
               run_per_partition=False):
    """
    Generate a datarun, including a dataset if necessary.

    sql_config: Object with all attributes necessary to initialize a Database.
    run_config: all attributes necessary to initialize a Datarun, including
        Dataset info if the dataset has not already been created.
    aws_config: all attributes necessary to connect to an S3 bucket.

    Returns: ID of the generated datarun
    """
    # connect to the database
    db = Database(sql_config.dialect, sql_config.database, sql_config.username,
                  sql_config.password, sql_config.host, sql_config.port,
                  sql_config.query)

    # if the user has provided a dataset id, use that. Otherwise, create a new
    # dataset based on the arguments we were passed.
    if run_config.dataset_id is None:
        dataset = create_dataset(db, run_config, aws_config=aws_config)
        run_config.dataset_id = dataset.id
    else:
        dataset = db.get_dataset(run_config.dataset_id)

    method_parts = {}
    for m in run_config.methods:
        # enumerate all combinations of categorical variables for this method
        method = Method(m)
        method_parts[m] = method.get_hyperpartitions()
        logger.info('method %s has %d hyperpartitions' %
                    (m, len(method_parts[m])))

    # create hyperpartitions and datarun(s)
    run_ids = []
    if not run_per_partition:
        logger.debug('saving datarun...')
        datarun = create_datarun(db, dataset, run_config)

    logger.debug('saving hyperpartions...')
    for method, parts in list(method_parts.items()):
        for part in parts:
            # if necessary, create a new datarun for each hyperpartition.
            # This setting is useful for debugging.
            if run_per_partition:
                datarun = create_datarun(db, dataset, run_config)
                run_ids.append(datarun.id)

            # create a new hyperpartition in the database
            db.create_hyperpartition(datarun_id=datarun.id,
                                     method=method,
                                     tunables=part.tunables,
                                     constants=part.constants,
                                     categoricals=part.categoricals,
                                     status=PartitionStatus.INCOMPLETE)

    logger.info('Data entry complete. Summary:')
    logger.info('\tDataset ID: %d' % dataset.id)
    logger.info('\tTraining data: %s' % dataset.train_path)
    logger.info('\tTest data: %s' % (dataset.test_path or 'None'))
    if run_per_partition:
        logger.info('\tDatarun IDs: %s' % ', '.join(map(str, run_ids)))
    else:
        logger.info('\tDatarun ID: %d' % datarun.id)
    logger.info('\tHyperpartition selection strategy: %s' % datarun.selector)
    logger.info('\tParameter tuning strategy: %s' % datarun.tuner)
    logger.info('\tBudget: %d (%s)' % (datarun.budget, datarun.budget_type))

    return run_ids or datarun.id



#!/usr/bin/python2.7
from __future__ import print_function
import argparse
import os
import yaml
from collections import defaultdict
from os.path import join

from atm.config import *
from atm.database import Database
from atm.enter_data import enter_data
from atm.utilities import download_file_s3
from atm.worker import work

from utilities import *


CONF_DIR = os.path.join(PROJECT_ROOT, 'config/test/')
DATA_DIR = os.path.join(PROJECT_ROOT, 'data/test/')
RUN_CONFIG = join(CONF_DIR, 'run-all.yaml')
SQL_CONFIG = join(CONF_DIR, 'sql-sqlite.yaml')

DATASETS_MAX_MIN = [
    'wholesale-customers_1.csv',
    'car_1.csv',
    'wall-robot-navigation_1.csv',
    'wall-robot-navigation_2.csv',
    'wall-robot-navigation_3.csv',
    'analcatdata_authorship_1.csv',
    'cardiotocography_1.csv',
    'wine_1.csv',
    'seismic-bumps_1.csv',
    'balance-scale_1.csv',
]
DATASETS_MAX_FIRST = [
    'wine_1.csv',
    'balance-scale_1.csv',
    'seeds_1.csv',
    'collins_1.csv',
    'cpu_1.csv',
    'vowel_1.csv',
    'car_2.csv',
    'hill-valley_2.csv',
    'rabe_97_1.csv',
    'monks-problems-2_1.csv',
]
DATASETS_SIMPLE = [
    'pollution_1.csv',  # binary test data
    'iris.data.csv',    # ternary test data
]

DATASETS = DATASETS_SIMPLE


parser = argparse.ArgumentParser(description='''
Run a single end-to-end test with 10 sample datasets.
The script will create a datarun for each dataset, then run a worker until the
jobs are finished.
''')
parser.add_argument('--processes', help='number of processes to run concurrently',
                    type=int, default=4)

args = parser.parse_args()
sql_config, run_config, _, _ = load_config(sql_path=SQL_CONFIG,
                                           run_path=RUN_CONFIG)

db = Database(**vars(sql_config))

print('creating dataruns...')
datarun_ids = []
for ds in DATASETS:
    run_config.train_path = join(DATA_DIR, ds)
    datarun_ids.append(enter_data(sql_config=sql_config,
                                  run_config=run_config))

work_parallel(db=db, datarun_ids=datarun_ids, n_procs=args.processes)

print('workers finished.')

for rid in datarun_ids:
    print_summary(db, rid)



from __future__ import print_function
import argparse
import numpy as np
import os

from collections import defaultdict
from multiprocessing import Process
from sklearn.metrics import auc

from atm.config import *
from atm.worker import work
from atm.database import db_session
from atm.utilities import download_file_http

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

BASELINE_PATH = os.path.join(PROJECT_ROOT, 'test/baselines/best_so_far_multi_trial/')
DATA_URL = 'https://s3.amazonaws.com/mit-dai-delphi-datastore/downloaded/'
BASELINE_URL = 'https://s3.amazonaws.com/mit-dai-delphi-datastore/best_so_far_multi_trial/'


def get_best_so_far(db, datarun_id):
    """
    Get a series representing best-so-far performance for datarun_id.
    """
    # generate a list of the "best so far" score after each classifier was
    # computed (in chronological order)
    classifiers = db.get_classifiers(datarun_id=datarun_id)
    y = []
    for l in classifiers:
        best_so_far = max(y + [l.cv_judgment_metric])
        y.append(best_so_far)
    return y


def graph_series(length, title, **series):
    """
    Graph series of performance metrics against one another.

    length: all series will be truncated to this length
    title: what to title the graph
    **series: mapping of labels to series of performance data
    """
    if plt is None:
        raise ImportError("Unable to import matplotlib")

    lines = []
    for label, data in series.items():
        # copy up to `length` of the values in `series` into y.
        y = data[:length]
        x = range(len(y))

        # plot y against x
        line, = plt.plot(x, y, '-', label=label)
        lines.append(line)

    plt.xlabel('classifiers')
    plt.ylabel('performance')
    plt.title(title)
    plt.legend(handles=lines)
    plt.show()

def report_auc_vs_baseline(db, rids, graph=False):

    if len(rids) == 0:
        return
    rid = rids[0]
    with db_session(db):
        run = db.get_datarun(rid)
        ds = run.dataset
        test = np.array([[float(y) for y in get_best_so_far(db, rid)] for rid in rids])
        test = test.T
        mean_test = np.mean(test, axis =1).tolist()

    ds_file = os.path.basename(ds.train_path)
    bl_path = download_file_http(BASELINE_URL + ds_file,
                                 local_folder=BASELINE_PATH)
    with open(bl_path) as f:
        baseline = np.array([[float(each) for each in l.strip().split('\t')] for l in f])
        mean_baseline = np.mean(baseline, axis =1).tolist()

    min_len = min(baseline.shape[0], test.shape[0])
    x = range(min_len)

    test_aucs = np.array([auc(x, test[:min_len,row]) for row in range(test.shape[1])])
    bl_aucs = np.array([auc(x, baseline[:min_len,row]) for row in range(baseline.shape[1])])
    # get avg, std, min of AUC over trials
    mean_auc_test = np.mean(test_aucs)
    mean_auc_bl = np.mean(bl_aucs)
    std_auc_test = np.std(test_aucs)
    std_auc_bl = np.std(bl_aucs)
    min_auc_test = np.min(test_aucs)
    min_auc_bl = np.min(bl_aucs)
    mean_auc_diff = mean_auc_test - mean_auc_bl
    print('Dataset %s (dataruns %s)' % (ds_file, rids))
    print ('Comparing %d trials to baseline generated by %d trials'%(len(rids), baseline.shape[1]))
    print('MEAN AUC: test = %.3f, baseline = %.3f (%.3f)' % (mean_auc_test, mean_auc_bl, mean_auc_diff))
    print('STD AUC: test = %.3f, baseline = %.3f' % (std_auc_test, std_auc_bl))
    print('MIN AUC: test = %.3f, baseline = %.3f' % (min_auc_test, min_auc_bl))

    if graph:
        graph_series(100, ds_file, baseline=mean_baseline, test=mean_test)

    return mean_auc_test, mean_auc_bl


def print_summary(db, rid):
    run = db.get_datarun(rid)
    ds = db.get_dataset(run.dataset_id)
    print()
    print('Dataset %s' % ds)
    print('Datarun %s' % run)

    classifiers = db.get_classifiers(datarun_id=rid)
    errs = db.get_classifiers(datarun_id=rid, status=ClassifierStatus.ERRORED)
    complete = db.get_classifiers(datarun_id=rid,
                                  status=ClassifierStatus.COMPLETE)
    print('Classifiers: %d total; %d errors, %d complete' %
          (len(classifiers), len(errs), len(complete)))

    best = db.get_best_classifier(score_target=run.score_target,
                                  datarun_id=run.id)
    if best is not None:
        score = best.cv_judgment_metric
        err = 2 * best.cv_judgment_metric_stdev
        print('Best result overall: classifier %d, %s = %.3f +- %.3f' %\
            (best.id, run.metric, score, err))


def print_method_summary(db, rid):
    # maps methods to sets of hyperpartitions, and hyperpartitions to lists of
    # classifiers
    alg_map = {a: defaultdict(list) for a in db.get_methods(datarun_id=rid)}

    run = db.get_datarun(rid)
    classifiers = db.get_classifiers(datarun_id=rid)
    for l in classifiers:
        hp = db.get_hyperpartition(l.hyperpartition_id)
        alg_map[hp.method][hp.id].append(l)

    for alg, hp_map in alg_map.items():
        print()
        print('method %s:' % alg)

        classifiers = sum(hp_map.values(), [])
        errored = len([l for l in classifiers if l.status ==
                       ClassifierStatus.ERRORED])
        complete = len([l for l in classifiers if l.status ==
                        ClassifierStatus.COMPLETE])
        print('\t%d errored, %d complete' % (errored, complete))

        best = db.get_best_classifier(score_target=run.score_target,
                                      datarun_id=rid, method=alg)
        if best is not None:
            score = best.cv_judgment_metric
            err = 2 * best.cv_judgment_metric_stdev
            print('\tBest: classifier %s, %s = %.3f +- %.3f' % (best, run.metric,
                                                                score, err))

def print_hp_summary(db, rid):
    run = db.get_datarun(rid)
    classifiers = db.get_classifiers(datarun_id=rid)

    part_map = defaultdict(list)
    for c in classifiers:
        hp = c.hyperpartition_id
        part_map[hp].append(c)

    for hp, classifiers in part_map.items():
        print()
        print('hyperpartition', hp)
        print(db.get_hyperpartition(hp))

        errored = len([c for c in classifiers if c.status ==
                       ClassifierStatus.ERRORED])
        complete = len([c for c in classifiers if c.status ==
                        ClassifierStatus.COMPLETE])
        print('\t%d errored, %d complete' % (errored, complete))

        best = db.get_best_classifier(score_target=run.score_target,
                                      datarun_id=rid, hyperpartition_id=hp)
        if best is not None:
            score = best.cv_judgment_metric
            err = 2 * best.cv_judgment_metric_stdev
            print('\tBest: classifier %s, %s = %.3f +- %.3f' % (best, run.metric,
                                                                score, err))

def work_parallel(db, datarun_ids=None, aws_config=None, n_procs=4):
    print('starting workers...')
    kwargs = dict(db=db, datarun_ids=datarun_ids, save_files=False,
                  choose_randomly=True, cloud_mode=False,
                  aws_config=aws_config, wait=False)

    if n_procs > 1:
        # spawn a set of worker processes to work on the dataruns
        procs = []
        for i in range(n_procs):
            p = Process(target=work, kwargs=kwargs)
            p.start()
            procs.append(p)

        # wait for them to finish
        for p in procs:
            p.join()
    else:
        work(**kwargs)



#!/usr/bin/python2.7
from __future__ import absolute_import, unicode_literals

import datetime
import imp
import logging
import os
import random
import time
import traceback
import warnings
from builtins import object, str
from collections import defaultdict
from operator import attrgetter

import numpy as np
from boto.s3.connection import Key as S3Key
from boto.s3.connection import S3Connection

from .config import LogConfig
from .constants import *
from .database import ClassifierStatus, db_session
from .model import Model
from .utilities import *

# shhh
warnings.filterwarnings('ignore')

# for garrays
os.environ['GNUMPY_IMPLICIT_CONVERSION'] = 'allow'

# how long to sleep between loops while waiting for new dataruns to be added
LOOP_WAIT = 1

# load the library-wide logger
logger = logging.getLogger('atm')


# Exception thrown when something goes wrong for the worker, but the worker
# handles the error.
class ClassifierError(Exception):
    pass


class Worker(object):
    def __init__(self, database, datarun, save_files=True, cloud_mode=False,
                 aws_config=None, log_config=None, public_ip='localhost'):
        """
        database: Database object with connection information
        datarun: Datarun ORM object to work on.
        save_files: if True, save model and metrics files to disk or cloud.
        cloud_mode: if True, save classifiers to the cloud
        aws_config: S3Config object with amazon s3 connection info
        """
        self.db = database
        self.datarun = datarun
        self.save_files = save_files
        self.cloud_mode = cloud_mode
        self.aws_config = aws_config
        self.public_ip = public_ip

        log_config = log_config or LogConfig()
        self.model_dir = log_config.model_dir
        self.metric_dir = log_config.metric_dir
        self.verbose_metrics = log_config.verbose_metrics
        ensure_directory(self.model_dir)
        ensure_directory(self.metric_dir)

        # load the Dataset from the database
        self.dataset = self.db.get_dataset(self.datarun.dataset_id)

        # load the Selector and Tuner classes specified by our datarun
        self.load_selector()
        self.load_tuner()

    def load_selector(self):
        """
        Load and initialize the BTB class which will be responsible for
        selecting hyperpartitions.
        """
        # selector will either be a key into SELECTORS_MAP or a path to
        # a file that defines a class called CustomSelector.
        if self.datarun.selector in SELECTORS_MAP:
            Selector = SELECTORS_MAP[self.datarun.selector]
        else:
            path, classname = re.match(CUSTOM_CLASS_REGEX,
                                       self.datarun.selector).groups()
            mod = imp.load_source('btb.selection.custom', path)
            Selector = getattr(mod, classname)
        logger.info('Selector: %s' % Selector)

        # generate the arguments we need to initialize the selector
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)
        hp_by_method = defaultdict(list)
        for hp in hyperpartitions:
            hp_by_method[hp.method].append(hp.id)
        hyperpartition_ids = [hp.id for hp in hyperpartitions]

        # Selector classes support passing in redundant arguments
        self.selector = Selector(choices=hyperpartition_ids,
                                 k=self.datarun.k_window,
                                 by_algorithm=dict(hp_by_method))

    def load_tuner(self):
        """
        Load, but don't initialize, the BTB class which will be responsible for
        choosing non-hyperpartition hyperparameter values (a subclass of Tuner). The
        tuner must be initialized with information about the hyperpartition, so it
        cannot be created until later.
        """
        # tuner will either be a key into TUNERS_MAP or a path to
        # a file that defines a class called CustomTuner.
        if self.datarun.tuner in TUNERS_MAP:
            self.Tuner = TUNERS_MAP[self.datarun.tuner]
        else:
            path, classname = re.match(CUSTOM_CLASS_REGEX,
                                       self.datarun.tuner).groups()
            mod = imp.load_source('btb.tuning.custom', path)
            self.Tuner = getattr(mod, classname)
        logger.info('Tuner: %s' % self.Tuner)

    def select_hyperpartition(self):
        """
        Use the hyperpartition selection method specified by our datarun to choose a
        hyperpartition of hyperparameters from the ModelHub. Only consider
        partitions for which gridding is not complete.
        """
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)

        # load classifiers and build scores lists
        # make sure all hyperpartitions are present in the dict, even ones that
        # don't have any classifiers. That way the selector can choose hyperpartitions
        # that haven't been scored yet.
        hyperpartition_scores = {fs.id: [] for fs in hyperpartitions}
        classifiers = self.db.get_classifiers(datarun_id=self.datarun.id)

        for c in classifiers:
            # ignore hyperpartitions for which gridding is done
            if c.hyperpartition_id not in hyperpartition_scores:
                continue

            # the cast to float is necessary because the score is a Decimal;
            # doing Decimal-float arithmetic throws errors later on.
            score = float(getattr(c, self.datarun.score_target) or 0)
            hyperpartition_scores[c.hyperpartition_id].append(score)

        hyperpartition_id = self.selector.select(hyperpartition_scores)
        return self.db.get_hyperpartition(hyperpartition_id)

    def tune_hyperparameters(self, hyperpartition):
        """
        Use the hyperparameter tuning method specified by our datarun to choose
        a set of hyperparameters from the potential space.
        """
        # Get parameter metadata for this hyperpartition
        tunables = hyperpartition.tunables

        # If there aren't any tunable parameters, we're done. Return the vector
        # of values in the hyperpartition and mark the set as finished.
        if not len(tunables):
            logger.warning('No tunables for hyperpartition %d' % hyperpartition.id)
            self.db.mark_hyperpartition_gridding_done(hyperpartition.id)
            return vector_to_params(vector=[],
                                    tunables=tunables,
                                    categoricals=hyperpartition.categoricals,
                                    constants=hyperpartition.constants)

        # Get previously-used parameters: every classifier should either be
        # completed or have thrown an error
        all_clfs = self.db.get_classifiers(hyperpartition_id=hyperpartition.id)
        classifiers = [c for c in all_clfs
                       if c.status == ClassifierStatus.COMPLETE]

        # Extract parameters and scores as numpy arrays from classifiers
        X = params_to_vectors([c.hyperparameter_values for c in classifiers],
                              tunables)
        y = np.array([float(getattr(c, self.datarun.score_target))
                      for c in classifiers])

        # Initialize the tuner and propose a new set of parameters
        # this has to be initialized with information from the hyperpartition, so we
        # need to do it fresh for each classifier (not in load_tuner)
        tuner = self.Tuner(tunables=tunables,
                           gridding=self.datarun.gridding,
                           r_minimum=self.datarun.r_minimum)
        tuner.fit(X, y)
        vector = tuner.propose()

        if vector is None and self.datarun.gridding:
            logger.info('Gridding done for hyperpartition %d' % hyperpartition.id)
            self.db.mark_hyperpartition_gridding_done(hyperpartition.id)
            return None

        # Convert the numpy array of parameters to a form that can be
        # interpreted by ATM, then return.
        return vector_to_params(vector=vector,
                                tunables=tunables,
                                categoricals=hyperpartition.categoricals,
                                constants=hyperpartition.constants)

    def test_classifier(self, method, params):
        """
        Given a set of fully-qualified hyperparameters, create and test a
        classifier model.
        Returns: Model object and metrics dictionary
        """
        model = Model(method=method, params=params,
                      judgment_metric=self.datarun.metric,
                      class_column=self.dataset.class_column,
                      verbose_metrics=self.verbose_metrics)
        train_path, test_path = download_data(self.dataset.train_path,
                                              self.dataset.test_path,
                                              self.aws_config)
        metrics = model.train_test(train_path=train_path,
                                   test_path=test_path)
        target = self.datarun.score_target

        def metric_string(model):
            if 'cv' in target or 'mu_sigma' in target:
                return '%.3f +- %.3f' % (model.cv_judgment_metric,
                                         2 * model.cv_judgment_metric_stdev)
            else:
                return '%.3f' % model.test_judgment_metric

        logger.info('Judgment metric (%s, %s): %s' % (self.datarun.metric,
                                                      target[:-len('_judgment_metric')],
                                                      metric_string(model)))

        old_best = self.db.get_best_classifier(datarun_id=self.datarun.id,
                                               score_target=target)
        if old_best is not None:
            if getattr(model, target) > getattr(old_best, target):
                logger.info('New best score! Previous best (classifier %s): %s'
                            % (old_best.id, metric_string(old_best)))
            else:
                logger.info('Best so far (classifier %s): %s' % (old_best.id,
                                                                 metric_string(old_best)))

        return model, metrics

    def save_classifier(self, classifier_id, model, metrics):
        """
        Update a classifier with metrics and model information and mark it as
        "complete"

        classifier_id: ID of the classifier to save
        model: Model object containing a serializable representation of the
            final model generated by this classifier.
        metrics: Dictionary containing cross-validation and test metrics data
            for the model.
        """
        # whether to save model and metrics data to the filesystem
        if self.save_files:
            # keep a database session open so that the utility functions can
            # access the linked hyperpartitions and dataruns
            with db_session(self.db):
                classifier = self.db.get_classifier(classifier_id)
                model_path = save_model(classifier, self.model_dir, model)
                metric_path = save_metrics(classifier, self.metric_dir, metrics)

            # if necessary, save model and metrics to Amazon S3 bucket
            if self.cloud_mode:
                try:
                    self.save_classifier_cloud(model_path, metric_path)
                except Exception:
                    msg = traceback.format_exc()
                    logger.error('Error in save_classifier_cloud()')
                    self.db.mark_classifier_errored(classifier_id,
                                                    error_message=msg)
        else:
            model_path = None
            metric_path = None

        # update the classifier in the database
        self.db.complete_classifier(classifier_id=classifier_id,
                                    model_location=model_path,
                                    metrics_location=metric_path,
                                    cv_score=model.cv_judgment_metric,
                                    cv_stdev=model.cv_judgment_metric_stdev,
                                    test_score=model.test_judgment_metric)

        # update this session's hyperpartition entry
        logger.info('Saved classifier %d.' % classifier_id)

    def save_classifier_cloud(self, local_model_path, local_metric_path):
        """
        Save a classifier to the S3 bucket supplied by aws_config. Saves a
        serialized representaion of the model as well as a detailed set
        of metrics.

        local_model_path: path to serialized model in the local file system
        local_metric_path: path to serialized metrics in the local file system
        """
        # TODO: This does not work
        conn = S3Connection(self.aws_config.access_key, self.aws_config.secret_key)
        bucket = conn.get_bucket(s3_bucket)

        if aws_folder:
            aws_model_path = os.path.join(aws_folder, local_model_path)
            aws_metric_path = os.path.join(aws_folder, local_metric_path)
        else:
            aws_model_path = local_model_path
            aws_metric_path = local_metric_path

        kmodel = S3Key(bucket)
        kmodel.key = aws_model_path
        kmodel.set_contents_from_filename(local_model_path)
        logger.info('Uploading model at %s to S3 bucket %s' % (s3_bucket,
                                                               local_model_path))

        kmodel = S3Key(bucket)
        kmodel.key = aws_metric_path
        kmodel.set_contents_from_filename(local_metric_path)
        logger.info('Uploading metrics at %s to S3 bucket %s' % (s3_bucket,
                                                                 local_metric_path))

        # delete the local copy of the model & metrics so that they don't fill
        # up the worker instance's hard drive
        logger.info('Deleting local copies of %s and %s' % (local_model_path,
                                                            local_metric_path))
        os.remove(local_model_path)
        os.remove(local_metric_path)

    def is_datarun_finished(self):
        """
        Check to see whether the datarun is finished. This could be due to the
        budget being exhausted or due to hyperparameter gridding being done.
        """
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)
        if not hyperpartitions:
            logger.warning('No incomplete hyperpartitions for datarun %d present in database.'
                           % self.datarun.id)
            return True

        if self.datarun.budget_type == 'classifier':
            # hyperpartition classifier counts are updated whenever a classifier
            # is created, so this will count running, errored, and complete.
            n_completed = len(self.db.get_classifiers(datarun_id=self.datarun.id))
            if n_completed >= self.datarun.budget:
                logger.warning('Classifier budget has run out!')
                return True

        elif self.datarun.budget_type == 'walltime':
            deadline = self.datarun.deadline
            if datetime.datetime.now() > deadline:
                logger.warning('Walltime budget has run out!')
                return True

        return False

    def run_classifier(self, hyperpartition_id=None):
        """
        Choose hyperparameters, then use them to test and save a Classifier.
        """
        # check to see if our work is done
        if self.is_datarun_finished():
            # marked the run as done successfully
            self.db.mark_datarun_complete(self.datarun.id)
            logger.warning('Datarun %d has ended.' % self.datarun.id)
            return

        try:
            logger.debug('Choosing hyperparameters...')
            if hyperpartition_id is not None:
                hyperpartition = self.db.get_hyperpartition(hyperpartition_id)
                if hyperpartition.datarun_id != self.datarun.id:
                    logger.error('Hyperpartition %d is not a part of datarun %d'
                                 % (hyperpartition_id, self.datarun.id))
                    return
            else:
                # use the multi-arm bandit to choose which hyperpartition to use next
                hyperpartition = self.select_hyperpartition()

            # use tuner to choose a set of parameters for the hyperpartition
            params = self.tune_hyperparameters(hyperpartition)
        except Exception:
            logger.error('Error choosing hyperparameters: datarun=%s' % str(self.datarun))
            logger.error(traceback.format_exc())
            raise ClassifierError()

        if params is None:
            logger.warning('No parameters chosen: hyperpartition %d is finished.'
                           % hyperpartition.id)
            return

        param_info = 'Chose parameters for method "%s":' % hyperpartition.method
        for k in sorted(params.keys()):
            param_info += '\n\t%s = %s' % (k, params[k])
        logger.info(param_info)

        logger.debug('Creating classifier...')
        classifier = self.db.start_classifier(hyperpartition_id=hyperpartition.id,
                                              datarun_id=self.datarun.id,
                                              host=self.public_ip,
                                              hyperparameter_values=params)

        try:
            logger.debug('Testing classifier...')
            model, metrics = self.test_classifier(hyperpartition.method, params)
            logger.debug('Saving classifier...')
            self.save_classifier(classifier.id, model, metrics)
        except Exception:
            msg = traceback.format_exc()
            logger.error('Error testing classifier: datarun=%s' % str(self.datarun))
            logger.error(msg)
            self.db.mark_classifier_errored(classifier.id, error_message=msg)
            raise ClassifierError()


def work(db, datarun_ids=None, save_files=False, choose_randomly=True,
         cloud_mode=False, aws_config=None, log_config=None, total_time=None,
         wait=True):
    """
    Check the ModelHub database for unfinished dataruns, and spawn workers to
    work on them as they are added. This process will continue to run until it
    exceeds total_time or is broken with ctrl-C.

    db: Database instance with which we can make queries to ModelHub
    datarun_ids (optional): list of IDs of dataruns to compute on. If None,
        this will work on all unfinished dataruns in the database.
    choose_randomly: if True, work on all highest-priority dataruns in random
        order. If False, work on them in sequential order (by ID)
    cloud_mode: if True, save processed datasets to AWS. If this option is set,
        aws_config must be supplied.
    aws_config (optional): if cloud_mode is set, this must be an AWSConfig
        object with connection details for an S3 bucket.
    total_time (optional): if set to an integer, this worker will only work for
        total_time seconds. Otherwise, it will continue working until all
        dataruns are complete (or indefinitely).
    wait: if True, once all dataruns in the database are complete, keep spinning
        and wait for new runs to be added. If False, exit once all dataruns are
        complete.
    """
    start_time = datetime.datetime.now()
    public_ip = get_public_ip()

    # main loop
    while True:
        # get all pending and running dataruns, or all pending/running dataruns
        # from the list we were given
        dataruns = db.get_dataruns(include_ids=datarun_ids,
                                   ignore_complete=True)
        if not dataruns:
            if wait:
                logger.warning('No dataruns found. Sleeping %d seconds and trying again.'
                               % LOOP_WAIT)
                time.sleep(LOOP_WAIT)
                continue
            else:
                logger.warning('No dataruns found. Exiting.')
                break

        max_priority = max([r.priority for r in dataruns])
        priority_runs = [r for r in dataruns if r.priority == max_priority]

        # either choose a run randomly, or take the run with the lowest ID
        if choose_randomly:
            run = random.choice(priority_runs)
        else:
            run = sorted(dataruns, key=attrgetter('id'))[0]

        # say we've started working on this datarun, if we haven't already
        db.mark_datarun_running(run.id)

        logger.info('Computing on datarun %d' % run.id)
        # actual work happens here
        worker = Worker(db, run, save_files=save_files,
                        cloud_mode=cloud_mode, aws_config=aws_config,
                        log_config=log_config, public_ip=public_ip)
        try:
            worker.run_classifier()
        except ClassifierError:
            # the exception has already been handled; just wait a sec so we
            # don't go out of control reporting errors
            logger.warning('Something went wrong. Sleeping %d seconds.' % LOOP_WAIT)
            time.sleep(LOOP_WAIT)

        elapsed_time = (datetime.datetime.now() - start_time).total_seconds()
        if total_time is not None and elapsed_time >= total_time:
            logger.warning('Total run time for worker exceeded; exiting.')
            break



import datetime
import os
import random

import mock
import numpy as np
import pytest
from mock import ANY, Mock, patch

import atm
from atm import PROJECT_ROOT
from atm.config import LogConfig, RunConfig, SQLConfig
from atm.constants import METRICS_BINARY, TIME_FMT
from atm.database import ClassifierStatus, Database, db_session
from atm.enter_data import enter_data
from atm.model import Model
from atm.utilities import download_data, load_metrics, load_model
from atm.worker import ClassifierError, Worker

from btb.selection import BestKReward, BestKVelocity, Selector
from btb.tuning import GP, GPEi, Tuner

DB_CACHE_PATH = os.path.join(PROJECT_ROOT, 'data/modelhub/test/')
DB_PATH = '/tmp/atm.db'
METRIC_DIR = '/tmp/metrics/'
MODEL_DIR = '/tmp/models/'

DATASET_ID = 1
DATARUN_ID = 2
HYPERPART_ID = 34
DT_PARAMS = {'criterion': 'gini', 'max_features': 0.5, 'max_depth': 3,
             'min_samples_split': 2, 'min_samples_leaf': 1}

# helper class to allow fuzzy arg matching
class StringWith(object):
    def __init__(self, match):
        self.match = match

    def __eq__(self, other):
        return self.match in other


# helper class to allow incomplete object matching
class ObjWithAttrs(object):
    def __init__(self, **kwargs):
        self.attrs = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, other):
        return all([getattr(other, k) == v for k, v in self.attrs.items()])

    def __repr__(self):
        return '<%s>' % ', '.join(['%s=%s' % i for i in self.attrs.items()])


@pytest.fixture
def db():
    os.remove(DB_PATH)
    db = Database(dialect='sqlite', database=DB_PATH)
    # load cached ModelHub state. This database snapshot has one dataset
    # (pollution_1.csv) and two dataruns, one complete and one with 33/100
    # classifiers finished.
    db.from_csv(DB_CACHE_PATH)
    return db


@pytest.fixture
def dataset(db):
    # return the only dataset
    return db.get_dataset(DATASET_ID)


@pytest.fixture
def datarun(db):
    # return the unfinished datarun
    return db.get_datarun(DATARUN_ID)


@pytest.fixture
def hyperpartition(db):
    # return a decision tree hyperpartition matching the static params above
    return db.get_hyperpartition(HYPERPART_ID)


@pytest.fixture
def model(dataset):
    train_path, _ = download_data(dataset.train_path)
    model = Model(method='dt', params=DT_PARAMS,
                  judgment_metric='roc_auc',
                  class_column=dataset.class_column)
    model.train_test(train_path=train_path)
    return model


@pytest.fixture
def metrics():
    cv_mets = [{k: random.random() for k in METRICS_BINARY} for i in range(5)]
    test_mets = {k: random.random() for k in METRICS_BINARY}
    return {'cv': cv_mets, 'test': test_mets}


@pytest.fixture
def worker(db, datarun):
    return Worker(db, datarun)


def get_new_worker(**kwargs):
    kwargs['methods'] = kwargs.get('methods', ['logreg', 'dt'])
    sql_conf = SQLConfig(database=DB_PATH)
    run_conf = RunConfig(**kwargs)
    run_id = enter_data(sql_conf, run_conf)
    db = Database(**vars(sql_conf))
    datarun = db.get_datarun(run_id)
    return Worker(db, datarun)


def test_load_selector_and_tuner(db, dataset):
    worker = get_new_worker(selector='bestkvel', k_window=7, tuner='gp')
    assert type(worker.selector) == BestKVelocity
    assert len(worker.selector.choices) == 8
    assert worker.selector.k == 7
    assert worker.Tuner == GP


def test_load_custom_selector_and_tuner(db, dataset):
    tuner_path = os.path.join(PROJECT_ROOT, 'tests/utilities/mytuner.py')
    selector_path = os.path.join(PROJECT_ROOT, 'tests/utilities/myselector.py')
    worker = get_new_worker(selector=selector_path + ':MySelector',
                            tuner=tuner_path + ':MyTuner')
    assert isinstance(worker.selector, Selector)
    assert issubclass(worker.Tuner, Tuner)


def test_select_hyperpartition(worker):
    """
    This won't test that BTB is working correctly, just that the ATM-BTB
    connection is working.
    """
    worker.db.get_hyperpartitions = Mock(return_value=[Mock(id=1)])
    clf_mock = Mock(hyperpartition_id=1, cv_judgment_metric=0.5)
    worker.db.get_classifiers = Mock(return_value=[clf_mock])
    worker.selector.select = Mock(return_value=1)

    hp = worker.select_hyperpartition()

    worker.selector.select.assert_called_with({1: [0.5]})
    assert hp.id == 1


def test_tune_hyperparameters(worker, hyperpartition):
    """
    This won't test that BTB is working correctly, just that the ATM-BTB
    connection is working.
    """
    mock_tuner = Mock()
    worker.Tuner = Mock(return_value=mock_tuner)

    with patch('atm.worker.vector_to_params') as vtp_mock:
        params = worker.tune_hyperparameters(hyperpartition)
        vtp_mock.assert_called()

    approximate_tunables = [(k, ObjWithAttrs(range=v.range))
                            for k, v in hyperpartition.tunables]
    worker.Tuner.assert_called_with(tunables=approximate_tunables,
                                    gridding=worker.datarun.gridding,
                                    r_minimum=worker.datarun.r_minimum)
    mock_tuner.fit.assert_called()
    mock_tuner.propose.assert_called()


def test_test_classifier(db, dataset):
    metric = 'roc_auc'
    worker = get_new_worker(metric=metric, score_target='mu_sigma')

    model, metrics = worker.test_classifier(method='dt', params=DT_PARAMS)
    judge_mets = [m[metric] for m in metrics['cv']]

    assert type(model) == Model
    assert model.judgment_metric == metric
    assert model.cv_judgment_metric == np.mean(judge_mets)
    assert model.cv_judgment_metric_stdev == np.std(judge_mets)


def test_save_classifier(db, datarun, model, metrics):
    log_conf = LogConfig(model_dir=MODEL_DIR, metric_dir=METRIC_DIR)
    worker = Worker(db, datarun, log_config=log_conf)
    hp = db.get_hyperpartitions(datarun_id=worker.datarun.id)[0]
    classifier = worker.db.start_classifier(hyperpartition_id=hp.id,
                                            datarun_id=worker.datarun.id,
                                            host='localhost',
                                            hyperparameter_values=DT_PARAMS)

    worker.db.complete_classifier = Mock()
    worker.save_classifier(classifier.id, model, metrics)
    worker.db.complete_classifier.assert_called()

    with db_session(worker.db):
        clf = db.get_classifier(classifier.id)

        loaded = load_model(clf, MODEL_DIR)
        assert type(loaded) == Model
        assert loaded.method == model.method
        assert loaded.random_state == model.random_state

        assert load_metrics(clf, METRIC_DIR) == metrics


def test_is_datarun_finished(db, dataset, datarun):
    r1 = db.get_datarun(1)
    worker = Worker(db, r1)
    assert worker.is_datarun_finished()

    r2 = db.get_datarun(2)
    worker = Worker(db, r2)
    assert not worker.is_datarun_finished()

    deadline = (datetime.datetime.now() - datetime.timedelta(seconds=1)).strftime(TIME_FMT)
    worker = get_new_worker(deadline=deadline)
    assert worker.is_datarun_finished()


def test_run_classifier(worker, hyperpartition, model, metrics):
    worker.select_hyperpartition = Mock(return_value=hyperpartition)
    worker.tune_hyperparameters = Mock(return_value=DT_PARAMS)
    worker.test_classifier = Mock(return_value=(model, metrics))
    worker.save_classifier = Mock()
    worker.db = Mock()

    # make sure the function shorts out if the datarun is finished
    worker.is_datarun_finished = Mock(return_value=True)
    worker.run_classifier()
    assert not worker.select_hyperpartition.called
    assert not worker.tune_hyperparameters.called

    # make sure things run smoothly: hyperparameters are chosen and a classifier
    # is created, tested, and saved.
    worker.is_datarun_finished = Mock(return_value=False)
    worker.run_classifier()
    worker.select_hyperpartition.assert_called_once()
    worker.tune_hyperparameters.assert_called_once_with(hyperpartition)
    worker.db.start_classifier.assert_called_once_with(hyperpartition_id=hyperpartition.id,
                                                       datarun_id=worker.datarun.id,
                                                       host=ANY,
                                                       hyperparameter_values=DT_PARAMS)
    worker.test_classifier.assert_called_once_with(hyperpartition.method, DT_PARAMS)
    worker.save_classifier.assert_called_once_with(ANY, model, metrics)

    # make sure hyperpartition specification works
    hp_id = hyperpartition.id + 1
    worker.db.get_hyperpartition = lambda i: ObjWithAttrs(id=i, method='dt',
                                                          datarun_id=worker.datarun.id)
    worker.run_classifier(hyperpartition_id=hp_id)
    worker.tune_hyperparameters.assert_called_with(ObjWithAttrs(id=hp_id))

    # make sure error handling works correctly
    worker.test_classifier.side_effect = ValueError('qwerty')
    with pytest.raises(ClassifierError):
        worker.run_classifier()
    worker.db.mark_classifier_errored.assert_called_with(
        ANY, error_message=StringWith('qwerty'))



from __future__ import absolute_import, unicode_literals

import logging
import os
import re
import socket
import sys
from argparse import ArgumentError, ArgumentTypeError, RawTextHelpFormatter
from builtins import map, object, str

import yaml

from .constants import *
from .utilities import ensure_directory


class Config(object):
    """
    Class which stores configuration for one aspect of ATM. Subclasses of
    Config should define the list of all configurable parameters and any
    default values for those parameters other than None (in PARAMETERS and
    DEFAULTS, respectively). The object can be initialized with any number of
    keyword arguments; only kwargs that are in PARAMETERS will be used. This
    means you can (relatively) safely do things like
        args = parser.parse_args()
        conf = Config(**vars(args))
    and only relevant parameters will be set.

    Subclasses do not need to define __init__ or any other methods.
    """
    # list of all parameters which may be set on this config object
    PARAMETERS = []
    # default values for all required parameters
    DEFAULTS = {}

    def __init__(self, **kwargs):
        for key in self.PARAMETERS:
            value = kwargs.get(key)

            # Here, if a keyword argument is set to None, it will be overridden
            # by the default value. AFAIK, this is the only way to deal with
            # keyword args passed in from argparse that weren't set on the
            # command line. That means you shouldn't define any PARAMETERS for
            # which None is a meaningful value; if you do, make sure None is
            # also the default.
            if key in self.DEFAULTS and value is None:
                value = self.DEFAULTS[key]

            setattr(self, key, value)


class AWSConfig(Config):
    """ Stores configuration for AWS S3 and EC2 connections """
    PARAMETERS = [
        # universal config
        'access_key',
        'secret_key',

        # s3 config
        's3_bucket',
        's3_folder',

        # ec2 config
        'ec2_region',
        'ec2_amis',
        'ec2_key_pair',
        'ec2_keyfile',
        'ec2_instance_type',
        'ec2_username',
        'num_instances',
        'num_workers_per_instance'
    ]

    DEFAULTS = {}


class SQLConfig(Config):
    """ Stores configuration for SQL database setup & connection """
    PARAMETERS = [
        'dialect',
        'database',
        'username',
        'password',
        'host',
        'port',
        'query'
    ]

    DEFAULTS = {
        'dialect': 'sqlite',
        'database': 'atm.db',
    }


class RunConfig(Config):
    """ Stores configuration for Dataset and Datarun setup """
    PARAMETERS = [
        # dataset config
        'train_path',
        'test_path',
        'data_description',
        'class_column',

        # datarun config
        'dataset_id',
        'methods',
        'priority',
        'budget_type',
        'budget',
        'deadline',
        'tuner',
        'r_minimum',
        'gridding',
        'selector',
        'k_window',
        'metric',
        'score_target'
    ]

    DEFAULTS = {
        'train_path': os.path.join(DATA_TEST_PATH, 'pollution_1.csv'),
        'class_column': 'class',
        'methods': ['logreg', 'dt', 'knn'],
        'priority': 1,
        'budget_type': 'classifier',
        'budget': 100,
        'tuner': 'uniform',
        'selector': 'uniform',
        'r_minimum': 2,
        'k_window': 3,
        'gridding': 0,
        'metric': 'f1',
        'score_target': 'cv',
    }


class LogConfig(Config):
    PARAMETERS = [
        'log_level_stdout',
        'log_level_file',
        'log_dir',
        'model_dir',
        'metric_dir',
        'verbose_metrics',
    ]

    DEFAULTS = {
        'log_level_stdout': 'ERROR',
        'log_level_file': 'INFO',
        'log_dir': 'logs',
        'model_dir': 'models',
        'metric_dir': 'metrics',
        'verbose_metrics': False,
    }


def initialize_logging(config):
    file_level = LOG_LEVELS.get(config.log_level_file.upper(),
                                logging.CRITICAL)
    stdout_level = LOG_LEVELS.get(config.log_level_stdout.upper(),
                                  logging.CRITICAL)

    handlers = []
    if file_level > logging.NOTSET:
        fmt = '%(asctime)-15s %(name)s - %(levelname)s  %(message)s'
        ensure_directory(config.log_dir)
        path = os.path.join(config.log_dir, socket.gethostname() + '.txt')
        handler = logging.FileHandler(path)
        handler.setFormatter(logging.Formatter(fmt))
        handler.setLevel(file_level)
        handlers.append(handler)

    if stdout_level > logging.NOTSET:
        fmt = '%(message)s'
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(fmt))
        handler.setLevel(stdout_level)
        handlers.append(handler)

    if not len(handlers):
        handlers.append(logging.NullHandler())

    for lib in ['atm', 'btb']:
        logger = logging.getLogger(lib)
        logger.setLevel(min(file_level, stdout_level))

        for h in logger.handlers:
            logger.removeHandler(h)

        for h in handlers:
            logger.addHandler(h)

        logger.propagate = False
        logger.debug('Logging is active for module %s.' % lib)


def option_or_path(options, regex=CUSTOM_CLASS_REGEX):
    def type_check(s):
        # first, check whether the argument is one of the preconfigured options
        if s in options:
            return s

        # otherwise, check it against the regex, and try to pull out a path to a
        # real file. The regex must extract the path to the file as groups()[0].
        match = re.match(regex, s)
        if match and os.path.isfile(match.groups()[0]):
            return s

        # if both of those fail, there's something wrong
        raise ArgumentTypeError('%s is not a valid option or path!' % s)

    return type_check


def add_arguments_logging(parser):
    """
    Add all argparse arguments needed to parse logging configuration from the
    command line.
    parser: an argparse.ArgumentParser object
    """
    # Config file path
    parser.add_argument('--log-config', help='path to yaml logging config file')

    # paths to saved files
    parser.add_argument('--model-dir',
                        help='Directory where computed models will be saved')
    parser.add_argument('--metric-dir',
                        help='Directory where model metrics will be saved')
    parser.add_argument('--log-dir',
                        help='Directory where logs will be saved')

    # hoe much information to log or save
    parser.add_argument('--verbose-metrics', action='store_true',
                        help='If set, compute full ROC and PR curves and '
                        'per-label metrics for each classifier')

    log_levels = list(map(str.lower, list(LOG_LEVELS.keys())))
    parser.add_argument('--log-level-file', choices=log_levels,
                        help='minimum log level to write to the log file')
    # if this is being called from the command line, print more information to
    # stdout by default
    parser.add_argument('--log-level-stdout', choices=log_levels,
                        help='minimum log level to write to stdout')

    return parser


def add_arguments_aws_s3(parser):
    """
    Add all argparse arguments needed to parse AWS S3 configuration from the
    command line. This is separate from aws_ec2 because usually only one set of
    arguments or the other is needed.
    parser: an argparse.ArgumentParser object
    """
    # Config file
    parser.add_argument('--aws-config', help='path to yaml AWS config file')

    # All of these arguments must start with --aws-, and must correspond to
    # keys present in the AWS config example file.
    # AWS API access key pair
    # try... catch because this might be called after aws_s3
    try:
        parser.add_argument('--aws-access-key', help='AWS access key')
        parser.add_argument('--aws-secret-key', help='AWS secret key')
    except ArgumentError:
        pass

    # S3-specific arguments
    parser.add_argument('--aws-s3-bucket', help='AWS S3 bucket to store data')
    parser.add_argument('--aws-s3-folder', help='Folder in AWS S3 bucket in which to store data')

    return parser


def add_arguments_aws_ec2(parser):
    """
    Add all argparse arguments needed to parse AWS EC2 configuration from the
    command line. This is separate from aws_s3 because usually only one set of
    arguments or the other is needed.
    parser: an argparse.ArgumentParser object
    """
    # Config file
    parser.add_argument('--aws-config', help='path to yaml AWS config file')

    # All of these arguments must start with --aws-, and must correspond to
    # keys present in the AWS config example file.
    # AWS API access key pair
    # try... catch because this might be called after aws_s3
    try:
        parser.add_argument('--aws-access-key', help='AWS access key')
        parser.add_argument('--aws-secret-key', help='AWS secret key')
    except ArgumentError:
        pass

    # AWS EC2 configurations
    parser.add_argument('--num-instances', help='Number of EC2 instances to start')
    parser.add_argument('--num-workers-per-instance', help='Number of ATM workers per instances')
    parser.add_argument('--ec2-region', help='Region to start instances in')
    parser.add_argument('--ec2-ami', help='Name of ATM AMI')
    parser.add_argument('--ec2-key-pair', help='AWS key pair to use for EC2 instances')
    parser.add_argument('--ec2-keyfile', help='Local path to key file (must match ec2-key-pair)')
    parser.add_argument('--ec2-instance-type', help='Type of EC2 instance to start')
    parser.add_argument('--ec2-username', help='Username to log into EC2 instance')

    return parser


def add_arguments_sql(parser):
    """
    Add all argparse arguments needed to parse configuration for the ModelHub
    SQL database from the command line.

    parser: an argparse.ArgumentParser object
    """
    # Config file
    parser.add_argument('--sql-config', help='path to yaml SQL config file')

    # All of these arguments must start with --sql-, and must correspond to
    # keys present in the SQL config example file.
    parser.add_argument('--sql-dialect', choices=SQL_DIALECTS,
                        help='Dialect of SQL to use')
    parser.add_argument('--sql-database',
                        help='Name of, or path to, SQL database')
    parser.add_argument('--sql-username', help='Username for SQL database')
    parser.add_argument('--sql-password', help='Password for SQL database')
    parser.add_argument('--sql-host', help='Hostname for database machine')
    parser.add_argument('--sql-port', help='Port used to connect to database')
    parser.add_argument('--sql-query', help='Specify extra login details')

    return parser


def add_arguments_datarun(parser):
    """
    Add all argparse arguments needed to parse dataset and datarun configuration
    from the command line.

    parser: an argparse.ArgumentParser object
    """
    # make sure the text for these arguments is formatted correctly
    # this allows newlines in the help strings
    parser.formatter_class = RawTextHelpFormatter

    # Config file
    parser.add_argument('--run-config', help='path to yaml datarun config file')

    ##  Dataset Arguments  #####################################################
    ############################################################################
    parser.add_argument('--dataset-id', type=int,
                        help="ID of dataset, if it's already in the database")

    # These are only relevant if dataset_id is not provided
    parser.add_argument('--train-path', help='Path to raw training data')
    parser.add_argument('--test-path', help='Path to raw test data (if applicable)')
    parser.add_argument('--data-description', help='Description of dataset')
    parser.add_argument('--class-column', help='Name of the class column in the input data')

    ##  Datarun Arguments  #####################################################
    ############################################################################
    # Notes:
    # - Support vector machines (svm) can take a long time to train. It's not an
    #   error, it's just part of what happens when the method happens to explore
    #   a crappy set of parameters on a powerful algo like this.
    # - Stochastic gradient descent (sgd) can sometimes fail on certain
    #   parameter settings as well. Don't worry, they train SUPER fast, and the
    #   worker.py will simply log the error and continue.
    #
    # Method options:
    #   logreg - logistic regression
    #   svm    - support vector machine
    #   sgd    - linear classifier with stochastic gradient descent
    #   dt     - decision tree
    #   et     - extra trees
    #   rf     - random forest
    #   gnb    - gaussian naive bayes
    #   mnb    - multinomial naive bayes
    #   bnb    - bernoulli naive bayes
    #   gp     - gaussian process
    #   pa     - passive aggressive
    #   knn    - K nearest neighbors
    #   mlp    - multi-layer perceptron
    parser.add_argument('--methods', nargs='+',
                        type=option_or_path(METHODS, JSON_REGEX),
                        help='Method or list of methods to use for '
                        'classification. Each method can either be one of the '
                        'pre-defined method codes listed below or a path to a '
                        'JSON file defining a custom method.' +
                        '\n\nOptions: [%s]' % ', '.join(str(s) for s in METHODS))
    parser.add_argument('--priority', type=int,
                        help='Priority of the datarun (higher = more important')
    parser.add_argument('--budget-type', choices=BUDGET_TYPES,
                        help='Type of budget to use')
    parser.add_argument('--budget', type=int,
                        help='Value of the budget, either in classifiers or minutes')
    parser.add_argument('--deadline',
                        help='Deadline for datarun completion. If provided, this '
                        'overrides the configured walltime budget.\nFormat: ' +
                        TIME_FMT.replace('%', '%%'))

    # Which field to use to judge performance, for the sake of AutoML
    # options:
    #   f1        - F1 score (harmonic mean of precision and recall)
    #   roc_auc   - area under the Receiver Operating Characteristic curve
    #   accuracy  - percent correct
    #   cohen_kappa     - measures accuracy, but controls for chance of guessing
    #                     correctly
    #   rank_accuracy   - multiclass only: percent of examples for which the true
    #                     label is in the top 1/3 most likely predicted labels
    #   ap        - average precision: nearly identical to area under
    #               precision/recall curve.
    #   mcc       - matthews correlation coefficient: good for unbalanced classes
    #
    # f1 and roc_auc may be appended with _micro or _macro to use with
    # multiclass problems.
    parser.add_argument('--metric', choices=METRICS,
                        help='Metric by which ATM should evaluate classifiers. '
                        'The metric function specified here will be used to '
                        'compute the "judgment metric" for each classifier.')

    # Which data to use for computing judgment score
    #   cv   - cross-validated performance on training data
    #   test - performance on test data
    #   mu_sigma - lower confidence bound on cv score
    parser.add_argument('--score-target', choices=SCORE_TARGETS,
                        help='Determines which judgment metric will be used to '
                        'search the hyperparameter space. "cv" will use the mean '
                        'cross-validated performance, "test" will use the '
                        'performance on a test dataset, and "mu_sigma" will use '
                        'the lower confidence bound on the CV performance.')

    ##  AutoML Arguments  ######################################################
    ############################################################################
    # hyperparameter selection strategy
    # How should ATM sample hyperparameters from a given hyperpartition?
    #    uniform  - pick randomly! (baseline)
    #    gp       - vanilla Gaussian Process
    #    gp_ei    - Gaussian Process expected improvement criterion
    #    gp_eivel - Gaussian Process expected improvement, with randomness added
    #               in based on velocity of improvement
    #   path to custom tuner, defined in python
    parser.add_argument('--tuner', type=option_or_path(TUNERS),
                        help='Type of BTB tuner to use. Can either be one of '
                        'the pre-configured tuners listed below or a path to a '
                        'custom tuner in the form "/path/to/tuner.py:ClassName".'
                        '\n\nOptions: [%s]' % ', '.join(str(s) for s in TUNERS))

    # How should ATM select a particular hyperpartition from the set of all
    # possible hyperpartitions?
    # Options:
    #   uniform      - pick randomly
    #   ucb1         - UCB1 multi-armed bandit
    #   bestk        - MAB using only the best K runs in each hyperpartition
    #   bestkvel     - MAB with velocity of best K runs
    #   purebestkvel - always return hyperpartition with highest velocity
    #   recentk      - MAB with most recent K runs
    #   recentkvel   - MAB with velocity of most recent K runs
    #   hieralg      - hierarchical MAB: choose a classifier first, then choose
    #                  a partition
    #   path to custom selector, defined in python
    parser.add_argument('--selector', type=option_or_path(SELECTORS),
                        help='Type of BTB selector to use. Can either be one of '
                        'the pre-configured selectors listed below or a path to a '
                        'custom tuner in the form "/path/to/selector.py:ClassName".'
                        '\n\nOptions: [%s]' % ', '.join(str(s) for s in SELECTORS))

    # r_minimum is the number of random runs performed in each hyperpartition before
    # allowing bayesian opt to select parameters. Consult the thesis to
    # understand what those mean, but essentially:
    #
    #  if (num_classifiers_trained_in_hyperpartition >= r_minimum)
    #    # train using sample criteria
    #  else
    #    # train using uniform (baseline)
    parser.add_argument('--r-minimum', type=int,
                        help='number of random runs to perform before tuning can occur')

    # k is number that xxx-k methods use. It is similar to r_minimum, except it is
    # called k_window and determines how much "history" ATM considers for certain
    # partition selection logics.
    parser.add_argument('--k-window', type=int,
                        help='number of previous scores considered by -k selector methods')

    # gridding determines whether or not sample selection will happen on a grid.
    # If any positive integer, a grid with `gridding` points on each axis is
    # established, and hyperparameter vectors are sampled from this finite
    # space. If 0 (or blank), hyperparameters are sampled from continuous
    # space, and there is no limit to the number of hyperparameter vectors that
    # may be tried.
    parser.add_argument('--gridding', type=int,
                        help='gridding factor (0: no gridding)')

    return parser


def load_config(sql_path=None, run_path=None, aws_path=None, log_path=None, **kwargs):
    """
    Load config objects from yaml files and command line arguments. Command line
    args override yaml files where applicable.

    Args:
        sql_path: path to .yaml file with SQL configuration
        run_path: path to .yaml file with Dataset and Datarun configuration
        aws_path: path to .yaml file with AWS configuration
        log_path: path to .yaml file with logging configuration
        **kwargs: miscellaneous arguments specifying individual configuration
            parameters. Any kwargs beginning with sql_ are SQL config
            arguments, any beginning with aws_ are AWS config.

    Returns: sql_conf, run_conf, aws_conf, log_conf
    """
    sql_args = {}
    run_args = {}
    aws_args = {}
    log_args = {}

    # kwargs are most likely generated by argparse.
    # Any unspecified argparse arguments will be None, so ignore those. We only
    # care about arguments explicitly specified by the user.
    kwargs = {k: v for k, v in list(kwargs.items()) if v is not None}

    # check the keyword args for config paths
    sql_path = sql_path or kwargs.get('sql_config')
    run_path = run_path or kwargs.get('run_config')
    aws_path = aws_path or kwargs.get('aws_config')
    log_path = log_path or kwargs.get('log_config')

    # load any yaml config files for which paths were provided
    if sql_path:
        with open(sql_path) as f:
            sql_args = yaml.load(f)

    if run_path:
        with open(run_path) as f:
            run_args = yaml.load(f)

    if aws_path:
        with open(aws_path) as f:
            aws_args = yaml.load(f)

    if log_path:
        with open(log_path) as f:
            log_args = yaml.load(f)

    # Use keyword args to override yaml config values
    sql_args.update({k.replace('sql_', ''): v for k, v in list(kwargs.items())
                     if 'sql_' in k})
    aws_args.update({k.replace('aws_', ''): v for k, v in list(kwargs.items())
                     if 'aws_' in k})
    run_args.update({k: v for k, v in list(kwargs.items()) if k in
                     RunConfig.PARAMETERS})
    log_args.update({k: v for k, v in list(kwargs.items()) if k in
                     LogConfig.PARAMETERS})

    # It's ok if there are some extra arguments that get passed in here; only
    # kwargs that correspond to real config values will be stored on the config
    # objects.
    sql_conf = SQLConfig(**sql_args)
    aws_conf = AWSConfig(**aws_args)
    run_conf = RunConfig(**run_args)
    log_conf = LogConfig(**log_args)

    return sql_conf, run_conf, aws_conf, log_conf



"""Auto Tune Models
A multi-user, multi-data AutoML framework.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import logging
import os

# Get the path of the project root, so that the rest of the project can
# reference files relative to there.
PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))

# this defines which modules will be imported by "from atm import *"
__all__ = ['config', 'constants', 'database', 'enter_data', 'method', 'metrics',
           'model', 'utilities', 'worker']

# by default, nothing should be logged
logger = logging.getLogger('atm')
logger.addHandler(logging.NullHandler())



#!/usr/bin/python2.7
import json

import pytest

from atm.method import Method


def test_enumerate():
    js = {'name': 'test', 'class': 'test'}
    js['hyperparameters'] = {
        'a': {'type': 'int_cat', 'values': [0, 3]},
        'b': {'type': 'int', 'range': [0, 3]},
        'c': {'type': 'bool', 'values': [True, False]},
        'd': {'type': 'string', 'values': ['x', 'y']},
        'e': {'type': 'float_cat', 'values': [-0.5, 0.5, 1.0]},
        'f': {'type': 'float', 'range': [0.5]},
        'g': {'type': 'list',
              'list_length': [1, 2, 3],
              'element': {'type': 'int_exp', 'range': [1e-3, 1e3]}}
    }
    js['root_hyperparameters'] = ['a', 'f']
    js['conditional_hyperparameters'] = {
        'a': {'0': ['b'], '3': ['c']},
        'c': {'True': ['d'], 'False': ['e', 'g']},
    }

    config_path = '/tmp/method.json'
    with open(config_path, 'w') as f:
        json.dump(js, f)

    hps = Method(config_path).get_hyperpartitions()

    assert len(hps) == 12
    assert all('a' in list(zip(*hp.categoricals))[0] for hp in hps)
    assert all(('f', 0.5) in hp.constants for hp in hps)
    assert len([hp for hp in hps if hp.tunables
                and 'b' in list(zip(*hp.tunables))[0]]) == 1



from __future__ import division, unicode_literals

from builtins import object

import numpy as np
import pandas as pd
from past.utils import old_div
from sklearn.preprocessing import LabelEncoder, OneHotEncoder


class MetaData(object):
    def __init__(self, class_column, train_path, test_path=None):
        """
        Compute a bunch of metadata about the dataset.

        class_column: name of dataframe column containing labels
        data_paths: paths to csvs with the same columns
        """
        data = pd.read_csv(train_path)
        if test_path is not None:
            data = data.append(pd.read_csv(test_path))

        # compute the portion of labels that are the most common value
        counts = data[class_column].value_counts()
        total_features = data.shape[1] - 1
        for c in data.columns:
            if data[c].dtype == 'object':
                total_features += len(np.unique(data[c])) - 1
        majority_percentage = old_div(float(max(counts)), float(sum(counts)))

        self.n_examples = data.shape[0]
        self.d_features = total_features
        self.k_classes = len(np.unique(data[class_column]))
        self.majority = majority_percentage
        self.size = np.array(data).nbytes


class DataEncoder(object):
    def __init__(self, class_column='class', feature_columns=None):
        self.class_column = class_column
        self.feature_columns = feature_columns

        # these will be trained with fit_encoders()
        self.column_encoders = {}
        self.feature_encoder = None
        self.label_encoder = None

    def fit(self, data):
        """
        Fit one-hot encoders for categorical features and an integer encoder for
        the label. These can be used later to transform raw data into a form
        that ATM can work with.

        data: pd.DataFrame of unprocessed data
        """
        if self.class_column not in data.columns:
            raise KeyError('Class column "%s" not found in dataset!' %
                           self.class_column)

        cat_cols = []
        if self.feature_columns is None:
            features = data.drop([self.class_column], axis=1)
            self.feature_columns = features.columns
        else:
            features = data[self.feature_columns]

        # encode categorical columns, leave ordinal values alone
        for column in features.columns:
            if features[column].dtype == 'object':
                # save the indices of categorical columns for one-hot encoding
                cat_cols.append(features.columns.get_loc(column))

                # encode each feature as an integer in range(unique_vals)
                le = LabelEncoder()
                features[column] = le.fit_transform(features[column])
                self.column_encoders[column] = le

        # One-hot encode the whole feature matrix.
        # Set sparse to False so that we can test for NaNs in the output
        self.feature_encoder = OneHotEncoder(categorical_features=cat_cols,
                                             sparse=False)
        self.feature_encoder.fit(features)

        # Train an encoder for the label as well
        labels = np.array(data[[self.class_column]])
        self.label_encoder = LabelEncoder()
        self.label_encoder.fit(labels)

    def transform(self, data):
        """
        Convert a DataFrame of labeled data to a feature matrix in the form
        that ATM can use.
        """
        if self.class_column in data:
            # pull labels into a separate series and transform them to integers
            labels = np.array(data[[self.class_column]])
            y = self.label_encoder.transform(labels)
            # drop the label column and transform the remaining features
        else:
            y = None

        features = data[self.feature_columns]

        # encode each categorical feature as an integer
        for column, encoder in list(self.column_encoders.items()):
            features[column] = encoder.transform(features[column])

        # one-hot encode the categorical features
        X = self.feature_encoder.transform(features)

        return X, y

    def fit_transform(self, data):
        """ Process data into a form that ATM can use. """
        self.fit(data)
        return self.transform(data)



from __future__ import print_function

import argparse
import os
import warnings

from atm import PROJECT_ROOT
from atm.config import (add_arguments_aws_s3, add_arguments_sql,
                        add_arguments_datarun, add_arguments_logging,
                        load_config, initialize_logging)
from atm.enter_data import enter_data

warnings.filterwarnings("ignore")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="""
Creates a dataset (if necessary) and a datarun and adds them to the ModelHub.
All required arguments have default values. Running this script with no
arguments will create a new dataset with the file in data/pollution_1.csv and a
new datarun with the default arguments listed below.

You can pass yaml configuration files (--sql-config, --aws-config, --run-config)
instead of passing individual arguments. Any arguments in the config files will
override arguments passed on the command line. See the examples in the config/
folder for more information. """)
    # Add argparse arguments for aws, sql, and datarun config
    add_arguments_aws_s3(parser)
    add_arguments_sql(parser)
    add_arguments_datarun(parser)
    add_arguments_logging(parser)
    parser.add_argument('--run-per-partition', default=False, action='store_true',
                        help='if set, generate a new datarun for each hyperpartition')

    args = parser.parse_args()

    # default logging config is different if initialized from the command line
    if args.log_config is None:
        args.log_config = os.path.join(PROJECT_ROOT,
                                       'config/templates/log-script.yaml')

    # create config objects from the config files and/or command line args
    sql_conf, run_conf, aws_conf, log_conf = load_config(sql_path=args.sql_config,
                                                         run_path=args.run_config,
                                                         aws_path=args.aws_config,
                                                         log_path=args.log_config,
                                                         **vars(args))
    initialize_logging(log_conf)

    # create and save the dataset and datarun
    enter_data(sql_conf, run_conf, aws_conf, args.run_per_partition)



from __future__ import absolute_import, unicode_literals

import base64
import hashlib
import json
import logging
import os
import pickle
import re
from builtins import str

import numpy as np
from boto.s3.connection import Key, S3Connection

from .constants import *

from btb import ParamTypes

from future import standard_library  # isort:skip
standard_library.install_aliases()

import urllib  # isort:skip


# global variable storing this machine's public IP address
# (so we only have to fetch it once)
public_ip = None

# URL which should give us our public-facing IP address
PUBLIC_IP_URL = 'http://ip.42.pl/raw'

logger = logging.getLogger('atm')


def hash_dict(dictionary, ignored_keys=None):
    """
    Hash a python dictionary to a hexadecimal string.
    http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
    """
    dictionary = dict(dictionary)  # copy dictionary
    for key in (ignored_keys or []):
        del dictionary[key]
    return hashlib.md5(repr(sorted(dictionary.items())).encode('utf8')).hexdigest()


def hash_nested_tuple(tup):
    """ Hash a nested tuple to hexadecimal """
    return hashlib.md5(repr(sorted(tup)).encode('utf8')).hexdigest()


def hash_string(s):
    """ Hash a string to hexadecimal """
    return hashlib.md5(str(s).encode('utf8')).hexdigest()


def ensure_directory(directory):
    """ Create directory if it doesn't exist. """
    if not os.path.exists(directory):
        os.makedirs(directory)


def get_public_ip():
    """
    Get the public IP address of this machine. If the request times out,
    return "localhost".
    """
    global public_ip
    if public_ip is None:
        try:
            response = urllib.request.urlopen(PUBLIC_IP_URL, timeout=2)
            data = str(response.read().strip())
            # pull an ip-looking set of numbers from the response
            match = re.search('\d+\.\d+\.\d+\.\d+', data)
            if match:
                public_ip = match.group()
        except Exception as e:  # any exception, doesn't matter what
            logger.error('could not get public IP: %s' % e)
            public_ip = 'localhost'

    return public_ip


def object_to_base_64(obj):
    """ Pickle and base64-encode an object. """
    pickled = pickle.dumps(obj)
    return base64.b64encode(pickled)


def base_64_to_object(b64str):
    """
    Inverse of object_to_base_64.
    Decode base64-encoded string and then unpickle it.
    """
    decoded = base64.b64decode(b64str)
    return pickle.loads(decoded)


def obj_has_method(obj, method):
    """http://stackoverflow.com/questions/34439/finding-what-methods-an-object-has"""
    return hasattr(obj, method) and callable(getattr(obj, method))


## Converting hyperparameters to and from BTB-compatible formats

def vector_to_params(vector, tunables, categoricals, constants):
    """
    Converts a numpy vector to a dictionary mapping keys to named parameters.

    vector: single example to convert

    Examples of the format for SVM sigmoid hyperpartition:

    tunables = [('C', HyperParameter(type='float_exp', range=(1e-05, 1e5))),
                ('degree', HyperParameter(type='int', range=(2, 4))),
                ('gamma', HyperParameter(type='float_exp', range=(1e-05, 1e5)))]

    categoricals = (('kernel', 'poly'),
                    ('probability', True),
                    ('_scale', True))

    constants = [('cache_size', 15000)]
    """
    params = {}

    # add the tunables
    for i, elt in enumerate(vector):
        key, struct = tunables[i]
        if struct.type in [ParamTypes.INT, ParamTypes.INT_EXP]:
            params[key] = int(elt)
        elif struct.type in [ParamTypes.FLOAT, ParamTypes.FLOAT_EXP]:
            params[key] = float(elt)
        else:
            raise ValueError('Unknown data type: {}'.format(struct.type))

    # add the fixed categorical settings and fixed constant values
    for key, value in categoricals + constants:
        params[key] = value

    return params


def params_to_vectors(params, tunables):
    """
    Converts a list of parameter vectors (with metadata) into a numpy array
    ready for BTB tuning.

    Args:
        params: list of hyperparameter vectors. Each vector is a dict mapping
            the names of parameters to those parameters' values.
        tunables: list of HyperParameter metadata structures describing all
            the optimizable hyperparameters that should be in each vector. e.g.

        tunables = [('C',  HyperParameter(type='float_exp', range=(1e-5, 1e5))),
                    ('degree',  HyperParameter('int', (2, 4))),
                    ('gamma',  HyperParameter('float_exp', (1e-05, 1e5)))]

    Returns:
        vectors: np.array of parameter vectors ready to be optimized by a
            Gaussian Process (or what have you).
            vectors.shape = (len(params), len(tunables))
    """
    # make sure params is iterable
    if not isinstance(params, (list, np.ndarray)):
        params = [params]

    keys = [k[0] for k in tunables]
    vectors = np.zeros((len(params), len(keys)))
    for i, p in enumerate(params):
        for j, k in enumerate(keys):
            vectors[i, j] = p[k]
    return vectors


## Serializing and deserializing data on disk

def _make_save_path_old(dir, classifier, suffix):
    """
    Generate the base save path for a classifier's model and metrics files,
    based on the classifier's dataset name and hyperparameters.
    """
    run_hash = hash_string(classifier.datarun.dataset.name)
    params_hash = hash_dict(classifier.hyperparameter_values)
    filename = "%s-%s-%s.%s" % (run_hash, params_hash,
                                classifier.datarun.description, suffix)
    return os.path.join(dir, filename)


def make_save_path(dir, classifier, suffix):
    """
    Generate the base save path for a classifier's model and metrics files,
    based on the classifier's dataset name and hyperparameters.
    """
    run_name = "".join([c for c in classifier.datarun.dataset.name
                        if c.isalnum() or c in (' ', '-', '_')]).rstrip()
    params_hash = hash_dict(classifier.hyperparameter_values)[:8]
    filename = "%s-%s.%s" % (run_name, params_hash, suffix)
    return os.path.join(dir, filename)


def save_model(classifier, model_dir, model):
    """
    Save a serialized version of a Model object for a particular classifier.
    The object will be stored at a path generated from the classifier's
    attributes.
    """
    path = make_save_path(model_dir, classifier, 'model')
    logger.info('Saving model in: %s' % path)
    with open(path, 'wb') as f:
        pickle.dump(model, f, protocol=pickle.HIGHEST_PROTOCOL)
    return path


def save_metrics(classifier, metric_dir, metrics):
    """
    Save a JSON-serialized version of a set of performance metrics for a
    particular classifier. The metrics will be stored at a path generated from
    the classifier's attributes.
    """
    path = make_save_path(metric_dir, classifier, 'metric')
    logger.info('Saving metrics in: %s' % path)
    with open(path, 'w') as f:
        json.dump(metrics, f)
    return path


def load_model(classifier, model_dir):
    """ Load the Model object for a particular classifier """
    path = make_save_path(model_dir, classifier, 'model')
    with open(path, 'rb') as f:
        return pickle.load(f)


def load_metrics(classifier, metric_dir):
    """ Load the performance metrics for a particular classifier """
    path = make_save_path(metric_dir, classifier, 'metric')
    with open(path) as f:
        return json.load(f)


## Downloading data from the web

def get_local_data_path(data_path):
    """
    given a data path of the form "s3://..." or "http://...", return the local
    path where the file should be stored once it's downloaded.
    """
    if data_path is None:
        return None, None

    m = re.match(S3_PREFIX, data_path)
    if m:
        path = data_path[len(m.group()):].split('/')
        return os.path.join(DATA_DL_PATH, path[-1]), FileType.S3

    m = re.match(HTTP_PREFIX, data_path)
    if m:
        path = data_path[len(m.group()):].split('/')
        return os.path.join(DATA_DL_PATH, path[-1]), FileType.HTTP

    return data_path, FileType.LOCAL


def download_file_s3(aws_path, aws_config, local_folder=DATA_DL_PATH):
    """ Download a file from an S3 bucket and save it in the local folder. """
    # remove the prefix and extract the S3 bucket, folder, and file name
    m = re.match(S3_PREFIX, aws_path)
    split = aws_path[len(m.group()):].split('/')
    s3_bucket = split.pop(0)
    s3_folder = '/'.join(split[:-1])
    keyname = split[-1]

    # create the local folder if necessary
    if local_folder is not None:
        ensure_directory(local_folder)
        path = os.path.join(local_folder, keyname)
    else:
        path = keyname

    if os.path.isfile(path):
        logger.warning('file %s already exists!' % path)
        return path

    conn = S3Connection(aws_config.access_key, aws_config.secret_key)
    bucket = conn.get_bucket(s3_bucket)

    if s3_folder:
        aws_keyname = os.path.join(s3_folder, keyname)
    else:
        aws_keyname = keyname

    logger.debug('downloading data from S3...')
    s3key = Key(bucket)
    s3key.key = aws_keyname
    s3key.get_contents_to_filename(path)
    logger.info('file saved at %s' % path)

    return path


def download_file_http(url, local_folder=DATA_DL_PATH):
    """ Download a file from a public URL and save it locally. """
    filename = url.split('/')[-1]
    if local_folder is not None:
        ensure_directory(local_folder)
        path = os.path.join(local_folder, filename)
    else:
        path = filename

    if os.path.isfile(path):
        logger.warning('file %s already exists!' % path)
        return path

    logger.debug('downloading data from %s...' % url)
    f = urllib.request.urlopen(url)
    data = f.read()
    with open(path, 'wb') as outfile:
        outfile.write(data)
    logger.info('file saved at %s' % path)

    return path


def download_data(train_path, test_path=None, aws_config=None):
    """
    Download a set of train/test data from AWS (if necessary) and return the
    path to the local data.
    """
    local_train_path, train_type = get_local_data_path(train_path)
    local_test_path, test_type = get_local_data_path(test_path)

    # if the data are not present locally, try to download them from the
    # internet (either an S3 bucket or a http connection)
    if not os.path.isfile(local_train_path):
        if train_type == FileType.HTTP:
            assert (download_file_http(train_path) == local_train_path)
        elif train_type == FileType.S3:
            assert (download_file_s3(train_path, aws_config=aws_config) ==
                    local_train_path)

    if local_test_path and not os.path.isfile(local_test_path):
        if test_type == FileType.HTTP:
            assert (download_file_http(test_path) == local_test_path)
        elif test_type == FileType.S3:
            assert (download_file_s3(test_path, aws_config=aws_config) ==
                    local_test_path)

    return local_train_path, local_test_path



from __future__ import absolute_import, unicode_literals

import json
import pickle
from builtins import object
from datetime import datetime
from operator import attrgetter

import pandas as pd
from sqlalchemy import (Column, DateTime, Enum, ForeignKey, Integer, MetaData,
                        Numeric, String, Text, and_, create_engine, func,
                        inspect)
from sqlalchemy.engine.url import URL
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.orm.properties import ColumnProperty

from .constants import *
from .utilities import *

# The maximum number of errors allowed in a single hyperpartition. If more than
# this many classifiers using a hyperpartition error, the hyperpartition will be
# considered broken and ignored for the rest of the datarun.
MAX_HYPERPARTITION_ERRORS = 3


def try_with_session(commit=False):
    """
    Decorator for instance methods on Database that need a sqlalchemy session.

    This wrapping function checks if the Database has an active session yet. If
    not, it wraps the function call in a `with db_session():` block.
    """
    def wrap(func):
        def call(db, *args, **kwargs):
            # if the Database has an active session, don't create a new one
            if db.session is not None:
                result = func(db, *args, **kwargs)
                if commit:
                    db.session.commit()
            else:
                # otherwise, use the session generator
                with db_session(db, commit=commit):
                    result = func(db, *args, **kwargs)

            return result
        return call
    return wrap


class db_session(object):
    def __init__(self, db, commit=False):
        self.db = db
        self.commit = commit

    def __enter__(self):
        self.db.session = self.db.get_session()

    def __exit__(self, type, error, traceback):
        if error is not None:
            self.db.session.rollback()
        elif self.commit:
            self.db.session.commit()

        self.db.session.close()
        self.db.session = None


class Database(object):
    def __init__(self, dialect, database, username=None, password=None,
                 host=None, port=None, query=None):
        """
        Accepts configuration for a database connection, and defines SQLAlchemy
        ORM objects for all the tables in the database.
        """
        db_url = URL(drivername=dialect, database=database, username=username,
                     password=password, host=host, port=port, query=query)
        self.engine = create_engine(db_url)
        self.session = None
        self.get_session = sessionmaker(bind=self.engine,
                                        expire_on_commit=False)

        # create ORM objects for the tables
        self._define_tables()

    def _define_tables(self):
        """
        Define the SQLAlchemy ORM class for each table in the ModelHub database.

        These must be defined after the Database class is initialized so that
        the database metadata is available (at runtime).
        If the database does not already exist, it will be created. If it does
        exist, it will not be updated with new schema -- after schema changes,
        the database must be destroyed and reinialized.
        """
        metadata = MetaData(bind=self.engine)
        Base = declarative_base(metadata=metadata)

        class Dataset(Base):
            __tablename__ = 'datasets'

            id = Column(Integer, primary_key=True, autoincrement=True)
            name = Column(String(100), nullable=False)

            # columns necessary for loading/processing data
            class_column = Column(String(100), nullable=False)
            train_path = Column(String(200), nullable=False)
            test_path = Column(String(200))
            description = Column(String(1000))

            # metadata columns, for convenience
            n_examples = Column(Integer, nullable=False)
            k_classes = Column(Integer, nullable=False)
            d_features = Column(Integer, nullable=False)
            majority = Column(Numeric(precision=10, scale=9), nullable=False)
            size_kb = Column(Integer, nullable=False)

            def __repr__(self):
                base = "<%s: %s, %d classes, %d features, %d rows>"
                return base % (self.name, self.description, self.k_classes,
                               self.d_features, self.n_examples)

        class Datarun(Base):
            __tablename__ = 'dataruns'

            # relational columns
            id = Column(Integer, primary_key=True, autoincrement=True)
            dataset_id = Column(Integer, ForeignKey('datasets.id'))
            dataset = relationship('Dataset', back_populates='dataruns')

            description = Column(String(200), nullable=False)
            priority = Column(Integer)

            # hyperparameter selection and tuning settings
            selector = Column(String(200), nullable=False)
            k_window = Column(Integer)
            tuner = Column(String(200), nullable=False)
            gridding = Column(Integer, nullable=False)
            r_minimum = Column(Integer)

            # budget settings
            budget_type = Column(Enum(*BUDGET_TYPES))
            budget = Column(Integer)
            deadline = Column(DateTime)

            # which metric to use for judgment, and how to compute it
            metric = Column(Enum(*METRICS))
            score_target = Column(Enum(*[s + '_judgment_metric' for s in
                                         SCORE_TARGETS]))

            # variables that store the status of the datarun
            start_time = Column(DateTime)
            end_time = Column(DateTime)
            status = Column(Enum(*DATARUN_STATUS), default=RunStatus.PENDING)

            def __repr__(self):
                base = "<ID = %d, dataset ID = %s, strategy = %s, budget = %s (%s), status: %s>"
                return base % (self.id, self.dataset_id, self.description,
                               self.budget_type, self.budget, self.status)

        Dataset.dataruns = relationship('Datarun', order_by='Datarun.id',
                                        back_populates='dataset')

        class Hyperpartition(Base):
            __tablename__ = 'hyperpartitions'

            # relational columns
            id = Column(Integer, primary_key=True, autoincrement=True)
            datarun_id = Column(Integer, ForeignKey('dataruns.id'))
            datarun = relationship('Datarun', back_populates='hyperpartitions')

            # name of or path to a configured classification method
            method = Column(String(255))

            # list of categorical parameters whose values are fixed to define
            # this hyperpartition
            categorical_hyperparameters_64 = Column(Text)

            # list of continuous parameters which are not fixed; their values
            # must be selected by a Tuner
            tunable_hyperparameters_64 = Column(Text)

            # list of categorical or continuous parameters whose values are
            # always fixed. These do not define the hyperpartition, but their
            # values must be passed on to the method. Here for convenience.
            constant_hyperparameters_64 = Column(Text)

            # has the partition had too many errors, or is gridding done?
            status = Column(Enum(*PARTITION_STATUS),
                            default=PartitionStatus.INCOMPLETE)

            @property
            def categoricals(self):
                """
                A list of categorical variables along with the fixed values
                which define this hyperpartition.
                Each element is a ('name', HyperParameter) tuple.
                """
                return base_64_to_object(self.categorical_hyperparameters_64)

            @categoricals.setter
            def categoricals(self, value):
                self.categorical_hyperparameters_64 = object_to_base_64(value)

            @property
            def tunables(self):
                """
                A list of parameters which are unspecified and must be selected
                with a Tuner. Each element is a ('name', HyperParameter) tuple.
                """
                return base_64_to_object(self.tunable_hyperparameters_64)

            @tunables.setter
            def tunables(self, value):
                self.tunable_hyperparameters_64 = object_to_base_64(value)

            @property
            def constants(self):
                return base_64_to_object(self.constant_hyperparameters_64)

            @constants.setter
            def constants(self, value):
                self.constant_hyperparameters_64 = object_to_base_64(value)

            def __repr__(self):
                return "<%s: %s>" % (self.method, self.categoricals)

        Datarun.hyperpartitions = relationship('Hyperpartition',
                                               order_by='Hyperpartition.id',
                                               back_populates='datarun')

        class Classifier(Base):
            __tablename__ = 'classifiers'

            # relational columns
            id = Column(Integer, primary_key=True, autoincrement=True)
            datarun_id = Column(Integer, ForeignKey('dataruns.id'))
            datarun = relationship('Datarun', back_populates='classifiers')
            hyperpartition_id = Column(Integer, ForeignKey('hyperpartitions.id'))
            hyperpartition = relationship('Hyperpartition',
                                          back_populates='classifiers')

            # name of the host where the model was trained
            host = Column(String(50))

            # these columns point to where the output is stored
            model_location = Column(String(300))
            metrics_location = Column(String(300))

            # base 64 encoding of the hyperparameter names and values
            hyperparameter_values_64 = Column(Text, nullable=False)

            # performance metrics
            cv_judgment_metric = Column(Numeric(precision=20, scale=10))
            cv_judgment_metric_stdev = Column(Numeric(precision=20, scale=10))
            test_judgment_metric = Column(Numeric(precision=20, scale=10))

            start_time = Column(DateTime)
            end_time = Column(DateTime)
            status = Column(Enum(*CLASSIFIER_STATUS), nullable=False)
            error_message = Column(Text)

            @property
            def hyperparameter_values(self):
                return base_64_to_object(self.hyperparameter_values_64)

            @hyperparameter_values.setter
            def hyperparameter_values(self, value):
                self.hyperparameter_values_64 = object_to_base_64(value)

            @property
            def mu_sigma_judgment_metric(self):
                # compute the lower confidence bound on the cross-validated
                # judgment metric
                if self.cv_judgment_metric is None:
                    return None
                return (self.cv_judgment_metric - 2 *
                        self.cv_judgment_metric_stdev)

            def __repr__(self):
                params = ', '.join(['%s: %s' % i for i in
                                    list(self.hyperparameter_values.items())])
                return "<id=%d, params=(%s)>" % (self.id, params)

        Datarun.classifiers = relationship('Classifier',
                                           order_by='Classifier.id',
                                           back_populates='datarun')
        Hyperpartition.classifiers = relationship('Classifier',
                                                  order_by='Classifier.id',
                                                  back_populates='hyperpartition')

        self.Dataset = Dataset
        self.Datarun = Datarun
        self.Hyperpartition = Hyperpartition
        self.Classifier = Classifier

        Base.metadata.create_all(bind=self.engine)

    ###########################################################################
    ##  Save/load the database  ###############################################
    ###########################################################################

    @try_with_session()
    def to_csv(self, path):
        """
        Save the entire ModelHub database as a set of CSVs in the given
        directory.
        """
        for table in ['datasets', 'dataruns', 'hyperpartitions', 'classifiers']:
            df = pd.read_sql('SELECT * FROM %s' % table, self.session.bind)
            df.to_csv(os.path.join(path, '%s.csv' % table), index=False)

    @try_with_session(commit=True)
    def from_csv(self, path):
        """
        Load a snapshot of the ModelHub database from a set of CSVs in the given
        directory.
        """
        for model, table in [(self.Dataset, 'dataset'),
                             (self.Datarun, 'datarun'),
                             (self.Hyperpartition, 'hyperpartition'),
                             (self.Classifier, 'classifier')]:
            df = pd.read_csv(os.path.join(path, '%ss.csv' % table))

            # parse datetime columns. This is necessary because SQLAlchemy can't
            # interpret strings as datetimes on its own.
            # yes, this is the easiest way to do it
            for c in inspect(model).attrs:
                if type(c) != ColumnProperty:
                    continue
                col = c.columns[0]
                if type(col.type) == DateTime:
                    df[c.key] = pd.to_datetime(df[c.key],
                                               infer_datetime_format=True)

            for _, r in df.iterrows():
                # replace NaN and NaT with None
                for k, v in list(r.iteritems()):
                    if pd.isnull(v):
                        r[k] = None

                # insert the row into the database
                create_func = getattr(self, 'create_%s' % table)
                create_func(**r)

    ###########################################################################
    ##  Standard query methods  ###############################################
    ###########################################################################

    @try_with_session()
    def get_dataset(self, dataset_id):
        """ Get a specific dataset. """
        return self.session.query(self.Dataset).get(dataset_id)

    @try_with_session()
    def get_datarun(self, datarun_id):
        """ Get a specific datarun. """
        return self.session.query(self.Datarun).get(datarun_id)

    @try_with_session()
    def get_dataruns(self, ignore_pending=False, ignore_running=False,
                     ignore_complete=True, include_ids=None, exclude_ids=None,
                     max_priority=True):
        """
        Get a list of all dataruns matching the chosen filters.

        Args:
            ignore_pending: if True, ignore dataruns that have not been started
            ignore_running: if True, ignore dataruns that are already running
            ignore_complete: if True, ignore completed dataruns
            include_ids: only include ids from this list
            exclude_ids: don't return any ids from this list
            max_priority: only return dataruns which have the highest priority
                of any in the filtered set
        """
        query = self.session.query(self.Datarun)
        if ignore_pending:
            query = query.filter(self.Datarun.status != RunStatus.PENDING)
        if ignore_running:
            query = query.filter(self.Datarun.status != RunStatus.RUNNING)
        if ignore_complete:
            query = query.filter(self.Datarun.status != RunStatus.COMPLETE)
        if include_ids:
            exclude_ids = exclude_ids or []
            ids = [i for i in include_ids if i not in exclude_ids]
            query = query.filter(self.Datarun.id.in_(ids))
        elif exclude_ids:
            query = query.filter(self.Datarun.id.notin_(exclude_ids))

        dataruns = query.all()

        if not len(dataruns):
            return None

        if max_priority:
            mp = max(dataruns, key=attrgetter('priority')).priority
            dataruns = [d for d in dataruns if d.priority == mp]

        return dataruns

    @try_with_session()
    def get_hyperpartition(self, hyperpartition_id):
        """ Get a specific classifier.  """
        return self.session.query(self.Hyperpartition).get(hyperpartition_id)

    @try_with_session()
    def get_hyperpartitions(self, dataset_id=None, datarun_id=None, method=None,
                            ignore_gridding_done=True, ignore_errored=True):
        """
        Return all the hyperpartitions in a given datarun by id.
        By default, only returns incomplete hyperpartitions.
        """
        query = self.session.query(self.Hyperpartition)
        if dataset_id is not None:
            query = query.join(self.Datarun)\
                .filter(self.Datarun.dataset_id == dataset_id)
        if datarun_id is not None:
            query = query.filter(self.Hyperpartition.datarun_id == datarun_id)
        if method is not None:
            query = query.filter(self.Hyperpartition.method == method)
        if ignore_gridding_done:
            query = query.filter(self.Hyperpartition.status !=
                                 PartitionStatus.GRIDDING_DONE)
        if ignore_errored:
            query = query.filter(self.Hyperpartition.status !=
                                 PartitionStatus.ERRORED)

        return query.all()

    @try_with_session()
    def get_classifier(self, classifier_id):
        """ Get a specific classifier. """
        return self.session.query(self.Classifier).get(classifier_id)

    @try_with_session()
    def get_classifiers(self, dataset_id=None, datarun_id=None, method=None,
                        hyperpartition_id=None, status=None):
        """ Get a set of classifiers, filtered by the passed-in arguments. """
        query = self.session.query(self.Classifier)
        if dataset_id is not None:
            query = query.join(self.Datarun)\
                .filter(self.Datarun.dataset_id == dataset_id)
        if datarun_id is not None:
            query = query.filter(self.Classifier.datarun_id == datarun_id)
        if method is not None:
            query = query.join(self.Hyperpartition)\
                .filter(self.Hyperpartition.method == method)
        if hyperpartition_id is not None:
            query = query.filter(self.Classifier.hyperpartition_id ==
                                 hyperpartition_id)
        if status is not None:
            query = query.filter(self.Classifier.status == status)

        return query.all()

    ###########################################################################
    ##  Special-purpose queries  ##############################################
    ###########################################################################

    @try_with_session()
    def is_datatun_gridding_done(self, datarun_id):
        """
        Check whether gridding is done for the entire datarun.
        """
        datarun = self.get_datarun(datarun_id)
        is_done = True
        for hyperpartition in datarun.hyperpartitions:
            # If any hyperpartiton has not finished gridding or errored out,
            # gridding is not done for the datarun.
            if hyperpartition.status == PartitionStatus.INCOMPLETE:
                is_done = False

        return is_done

    @try_with_session()
    def get_number_of_hyperpartition_errors(self, hyperpartition_id):
        """
        Get the number of classifiers that have errored using a specified
        hyperpartition.
        """
        classifiers = self.session.query(self.Classifier)\
            .filter(and_(self.Classifier.hyperpartition_id == hyperpartition_id,
                         self.Classifier.status == ClassifierStatus.ERRORED)).all()
        return len(classifiers)

    @try_with_session()
    def get_methods(self, dataset_id=None, datarun_id=None,
                    ignore_errored=False, ignore_gridding_done=False):
        """ Get all methods used in a particular datarun. """
        hyperpartitions = self.get_hyperpartitions(dataset_id=dataset_id,
                                                   datarun_id=datarun_id,
                                                   ignore_gridding_done=False,
                                                   ignore_errored=False)
        methods = set(f.method for f in hyperpartitions)
        return list(methods)

    @try_with_session()
    def get_maximum_y(self, datarun_id, score_target):
        """ Get the maximum value of a numeric column by name, or None. """
        query = self.session.query(func.max(getattr(self.Classifier,
                                                    score_target)))
        result = query.filter(self.Classifier.datarun_id == datarun_id).first()
        if result:
            return float(result)
        return None

    @try_with_session()
    def get_best_classifier(self, score_target, dataset_id=None,
                            datarun_id=None, method=None,
                            hyperpartition_id=None):
        """
        Get the classifier with the best judgment metric, as indicated by
        score_target.

        score_target: indicates the metric by which to judge the best classifier.
        """
        classifiers = self.get_classifiers(dataset_id=dataset_id,
                                           datarun_id=datarun_id,
                                           method=method,
                                           hyperpartition_id=hyperpartition_id,
                                           status=ClassifierStatus.COMPLETE)

        if '_judgment_metric' not in score_target:
            score_target += '_judgment_metric'

        if not classifiers:
            return None
        return max(classifiers, key=attrgetter(score_target))

    @try_with_session()
    def load_model(self, classifier_id):
        clf = self.get_classifier(classifier_id)
        with open(clf.model_location, 'rb') as f:
            return pickle.load(f)

    @try_with_session()
    def load_metrics(self, classifier_id):
        clf = self.get_classifier(classifier_id)
        with open(clf.metrics_location, 'r') as f:
            return json.load(f)

    ###########################################################################
    ##  Methods to update the database  #######################################
    ###########################################################################

    @try_with_session(commit=True)
    def create_dataset(self, **kwargs):
        dataset = self.Dataset(**kwargs)
        self.session.add(dataset)
        return dataset

    @try_with_session(commit=True)
    def create_datarun(self, **kwargs):
        datarun = self.Datarun(**kwargs)
        self.session.add(datarun)
        return datarun

    @try_with_session(commit=True)
    def create_hyperpartition(self, **kwargs):
        partition = self.Hyperpartition(**kwargs)
        self.session.add(partition)
        return partition

    @try_with_session(commit=True)
    def create_classifier(self, **kwargs):
        classifier = self.Classifier(**kwargs)
        self.session.add(classifier)
        return classifier

    @try_with_session(commit=True)
    def start_classifier(self, hyperpartition_id, datarun_id, host,
                         hyperparameter_values):
        """
        Save a new, fully qualified classifier object to the database.
        Returns: the ID of the newly-created classifier
        """
        classifier = self.Classifier(hyperpartition_id=hyperpartition_id,
                                     datarun_id=datarun_id,
                                     host=host,
                                     hyperparameter_values=hyperparameter_values,
                                     start_time=datetime.now(),
                                     status=ClassifierStatus.RUNNING)
        self.session.add(classifier)
        return classifier

    @try_with_session(commit=True)
    def complete_classifier(self, classifier_id, model_location,
                            metrics_location, cv_score, cv_stdev, test_score):
        """
        Set all the parameters on a classifier that haven't yet been set, and mark
        it as complete.
        """
        classifier = self.session.query(self.Classifier).get(classifier_id)

        classifier.model_location = model_location
        classifier.metrics_location = metrics_location
        classifier.cv_judgment_metric = cv_score
        classifier.cv_judgment_metric_stdev = cv_stdev
        classifier.test_judgment_metric = test_score
        classifier.end_time = datetime.now()
        classifier.status = ClassifierStatus.COMPLETE

    @try_with_session(commit=True)
    def mark_classifier_errored(self, classifier_id, error_message):
        """
        Mark an existing classifier as having errored and set the error message. If
        the classifier's hyperpartiton has produced too many erring classifiers, mark it
        as errored as well.
        """
        classifier = self.session.query(self.Classifier).get(classifier_id)
        classifier.error_message = error_message
        classifier.status = ClassifierStatus.ERRORED
        classifier.end_time = datetime.now()
        if (self.get_number_of_hyperpartition_errors(classifier.hyperpartition_id) >
                MAX_HYPERPARTITION_ERRORS):
            self.mark_hyperpartition_errored(classifier.hyperpartition_id)

    @try_with_session(commit=True)
    def mark_hyperpartition_gridding_done(self, hyperpartition_id):
        """
        Mark a hyperpartiton as having all of its possible grid points explored.
        """
        hyperpartition = self.get_hyperpartition(hyperpartition_id)
        hyperpartition.status = PartitionStatus.GRIDDING_DONE

    @try_with_session(commit=True)
    def mark_hyperpartition_errored(self, hyperpartition_id):
        """
        Mark a hyperpartiton as having had too many classifier errors. This will
        prevent more classifiers from being trained on this hyperpartiton in the
        future.
        """
        hyperpartition = self.get_hyperpartition(hyperpartition_id)
        hyperpartition.status = PartitionStatus.ERRORED

    @try_with_session(commit=True)
    def mark_datarun_running(self, datarun_id):
        """
        Set the status of the Datarun to RUNNING and set the 'start_time' field
        to the current datetime.
        """
        datarun = self.get_datarun(datarun_id)
        if datarun.status == RunStatus.PENDING:
            datarun.status = RunStatus.RUNNING
            datarun.start_time = datetime.now()

    @try_with_session(commit=True)
    def mark_datarun_complete(self, datarun_id):
        """
        Set the status of the Datarun to COMPLETE and set the 'end_time' field
        to the current datetime.
        """
        datarun = self.get_datarun(datarun_id)
        datarun.status = RunStatus.COMPLETE
        datarun.end_time = datetime.now()



from __future__ import absolute_import, unicode_literals

import logging
import os
from builtins import object

from . import PROJECT_ROOT

from btb.selection import Uniform as UniformSelector
from btb.selection import (UCB1, BestKReward, BestKVelocity,
                           HierarchicalByAlgorithm, PureBestKVelocity,
                           RecentKReward, RecentKVelocity)
from btb.tuning import Uniform as UniformTuner
from btb.tuning import GP, GPEi, GPEiVelocity

# A bunch of constants which are used throughout the project, mostly for config.
# TODO: convert these lists and classes to something more elegant, like enums
SQL_DIALECTS = ['sqlite', 'mysql']
SCORE_TARGETS = ['cv', 'test', 'mu_sigma']
BUDGET_TYPES = ['none', 'classifier', 'walltime']
METHODS = ['logreg', 'svm', 'sgd', 'dt', 'et', 'rf', 'gnb', 'mnb', 'bnb',
           'gp', 'pa', 'knn', 'mlp', 'ada']
TUNERS = ['uniform', 'gp', 'gp_ei', 'gp_eivel']
SELECTORS = ['uniform', 'ucb1', 'bestk', 'bestkvel', 'purebestkvel', 'recentk',
             'recentkvel', 'hieralg']
DATARUN_STATUS = ['pending', 'running', 'complete']
CLASSIFIER_STATUS = ['running', 'errored', 'complete']
PARTITION_STATUS = ['incomplete', 'errored', 'gridding_done']

S3_PREFIX = '^s3://'
HTTP_PREFIX = '^https?://'

TIME_FMT = '%Y-%m-%d %H:%M'
DATA_TEST_PATH = os.path.join(PROJECT_ROOT, 'data/test')
DATA_DL_PATH = os.path.join(PROJECT_ROOT, 'data/downloads')
METHOD_PATH = os.path.join(PROJECT_ROOT, 'methods')

CUSTOM_CLASS_REGEX = '(.*\.py):(\w+)$'
JSON_REGEX = '(.*\.json)$'

N_FOLDS_DEFAULT = 10

LOG_LEVELS = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'NONE': logging.NOTSET
}

TUNERS_MAP = {
    'uniform': UniformTuner,
    'gp': GP,
    'gp_ei': GPEi,
    'gp_eivel': GPEiVelocity,
}

SELECTORS_MAP = {
    'uniform': UniformSelector,
    'ucb1': UCB1,
    'bestk': BestKReward,
    'bestkvel': BestKVelocity,
    'purebestkvel': PureBestKVelocity,
    'recentk': RecentKReward,
    'recentkvel': RecentKVelocity,
    'hieralg': HierarchicalByAlgorithm,
}

METHODS_MAP = {
    'logreg': 'logistic_regression.json',
    'svm': 'support_vector_machine.json',
    'sgd': 'stochastic_gradient_descent.json',
    'dt': 'decision_tree.json',
    'et': 'extra_trees.json',
    'rf': 'random_forest.json',
    'gnb': 'gaussian_naive_bayes.json',
    'mnb': 'multinomial_naive_bayes.json',
    'bnb': 'bernoulli_naive_bayes.json',
    'gp': 'gaussian_process.json',
    'pa': 'passive_aggressive.json',
    'knn': 'k_nearest_neighbors.json',
    'mlp': 'multi_layer_perceptron.json',
    'ada': 'adaboost.json'
}


class ClassifierStatus(object):
    RUNNING = 'running'
    ERRORED = 'errored'
    COMPLETE = 'complete'


class RunStatus(object):
    PENDING = 'pending'
    RUNNING = 'running'
    COMPLETE = 'complete'


class PartitionStatus(object):
    INCOMPLETE = 'incomplete'
    GRIDDING_DONE = 'gridding_done'
    ERRORED = 'errored'


class FileType(object):
    LOCAL = 'local'
    S3 = 's3'
    HTTP = 'http'


# these are the strings that are used to index into results dictionaries
class Metrics(object):
    ACCURACY = 'accuracy'
    RANK_ACCURACY = 'rank_accuracy'
    COHEN_KAPPA = 'cohen_kappa'
    F1 = 'f1'
    F1_MICRO = 'f1_micro'
    F1_MACRO = 'f1_macro'
    ROC_AUC = 'roc_auc'     # receiver operating characteristic
    ROC_AUC_MICRO = 'roc_auc_micro'
    ROC_AUC_MACRO = 'roc_auc_macro'
    AP = 'ap'               # average precision
    MCC = 'mcc'             # matthews correlation coefficient
    PR_CURVE = 'pr_curve'
    ROC_CURVE = 'roc_curve'


METRICS_BINARY = [
    Metrics.ACCURACY,
    Metrics.COHEN_KAPPA,
    Metrics.F1,
    Metrics.ROC_AUC,
    Metrics.AP,
    Metrics.MCC,
]

METRICS_MULTICLASS = [
    Metrics.ACCURACY,
    Metrics.RANK_ACCURACY,
    Metrics.COHEN_KAPPA,
    Metrics.F1_MICRO,
    Metrics.F1_MACRO,
    Metrics.ROC_AUC_MICRO,
    Metrics.ROC_AUC_MACRO,
]

METRICS = list(set(METRICS_BINARY + METRICS_MULTICLASS))



from __future__ import print_function
import argparse
import os
import random
from os.path import join

from atm import PROJECT_ROOT
from atm.config import *
from atm.database import Database
from atm.enter_data import enter_data

from utilities import *


CONF_DIR = os.path.join(PROJECT_ROOT, 'config/test/')
RUN_CONFIG = join(CONF_DIR, 'run-default.yaml')
SQL_CONFIG = join(CONF_DIR, 'sql-sqlite.yaml')

DATASETS_MAX_FIRST = [
    'collins_1.csv',
    'cpu_1.csv',
    'vowel_1.csv',
    'car_2.csv',
    'hill-valley_2.csv',
    'rabe_97_1.csv',
    'monks-problems-2_1.csv',
    # these datasets do not have baseline numbers
    #'wine_1.csv',
    #'balance-scale_1.csv',
    #'seeds_1.csv',
]


def btb_test(dataruns=None, datasets=None, processes=1, graph=False, **kwargs):
    """
    Run a test datarun using the chosen tuner and selector, and compare it to
    the baseline performance.

    Tuner and selector will be specified in **kwargs, along with the rest of the
    standard datarun arguments.
    """
    sql_conf, run_conf, _, _ = load_config(sql_path=SQL_CONFIG,
                                           run_path=RUN_CONFIG,
                                           **kwargs)

    db = Database(**vars(sql_conf))
    datarun_ids = dataruns or []
    datarun_ids_per_dataset = [[each] for each in dataruns] if dataruns else []
    datasets = datasets or DATASETS_MAX_FIRST

    # if necessary, generate datasets and dataruns
    if not datarun_ids:
        for ds in datasets:
            run_conf.train_path = DATA_URL + ds
            run_conf.dataset_id = None
            print('Creating 10 dataruns for', run_conf.train_path)
            run_ids = [enter_data(sql_conf, run_conf) for i in range(10)]
            datarun_ids_per_dataset.append(run_ids)
            datarun_ids.extend(run_ids)

    # work on the dataruns til they're done
    print('Working on %d dataruns' % len(datarun_ids))
    work_parallel(db=db, datarun_ids=datarun_ids, n_procs=processes)
    print('Finished!')

    results = {}

    # compute and maybe graph the results for each dataset
    for rids in datarun_ids_per_dataset:
        res = report_auc_vs_baseline(db, rids, graph=graph)
        results[tuple(rids)] = {'test': res[0], 'baseline': res[1]}

    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='''
    Test the performance of an AutoML method and compare it to the baseline
    performance curve.
    ''')
    parser.add_argument('--processes', help='number of processes to run concurrently',
                        type=int, default=1)
    parser.add_argument('--graph', action='store_true', default=False,
                        help='if this flag is inculded, graph the best-so-far '
                        'results of each datarun against the baseline.')
    parser.add_argument('--dataruns', nargs='+', type=int,
                        help='(optional) IDs of previously-created dataruns to '
                        'graph. If this option is included, no new dataruns '
                        'will be created, but any of the specified dataruns '
                        'will be finished if they are not already.')
    parser.add_argument('--datasets', nargs='+',
                        help='(optional) file names of training data to use. '
                        'Each should be a csv file present in the downloaded/ '
                        'folder of the HDI project S3 bucket '
                        '(https://s3.amazonaws.com/mit-dai-delphi-datastore/downloaded/).'
                        'The default is to use the files in DATASETS_MAX_FIRST.')
    add_arguments_datarun(parser)
    args = parser.parse_args()

    btb_test(**vars(args))



import json
import os

import pytest

from atm import PROJECT_ROOT, constants
from atm.config import RunConfig, SQLConfig
from atm.database import Database, db_session
from atm.enter_data import create_datarun, create_dataset, enter_data
from atm.utilities import get_local_data_path

DB_PATH = '/tmp/atm.db'
DB_CACHE_PATH = os.path.join(PROJECT_ROOT, 'data/modelhub/test/')
DATA_URL = 'https://s3.amazonaws.com/mit-dai-delphi-datastore/downloaded/'
BASELINE_PATH = os.path.join(PROJECT_ROOT, 'data/baselines/best_so_far/')
BASELINE_URL = 'https://s3.amazonaws.com/mit-dai-delphi-datastore/best_so_far/'

METHOD_HYPERPARTS = {
    'logreg': 6,
    'svm': 4,
    'sgd': 24,
    'dt': 2,
    'et': 2,
    'rf': 2,
    'gnb': 1,
    'mnb': 1,
    'bnb': 1,
    'gp': 8,
    'pa': 4,
    'knn': 24,
    'mlp': 60,
}


@pytest.fixture
def db():
    if os.path.exists(DB_PATH):
        os.remove(DB_PATH)
    db = Database(dialect='sqlite', database=DB_PATH)
    # load cached ModelHub state. This database snapshot has one dataset
    # (pollution_1.csv) and two dataruns, one complete and one with 33/100
    # classifiers finished.
    db.from_csv(DB_CACHE_PATH)
    return db


@pytest.fixture
def dataset(db):
    return db.get_dataset(1)


def test_create_dataset(db):
    train_url = DATA_URL + 'pollution_1_train.csv'
    test_url = DATA_URL + 'pollution_1_test.csv'

    train_path_local, _ = get_local_data_path(train_url)
    if os.path.exists(train_path_local):
        os.remove(train_path_local)

    test_path_local, _ = get_local_data_path(test_url)
    if os.path.exists(test_path_local):
        os.remove(test_path_local)

    run_conf = RunConfig(train_path=train_url,
                         test_path=test_url,
                         data_description='test',
                         class_column='class')
    dataset = create_dataset(db, run_conf)
    dataset = db.get_dataset(dataset.id)

    assert os.path.exists(train_path_local)
    assert os.path.exists(test_path_local)

    assert dataset.train_path == train_url
    assert dataset.test_path == test_url
    assert dataset.description == 'test'
    assert dataset.class_column == 'class'
    assert dataset.n_examples == 60
    assert dataset.d_features == 16
    assert dataset.k_classes == 2
    assert dataset.majority >= 0.5


def test_enter_data_by_methods(dataset):
    sql_conf = SQLConfig(database=DB_PATH)
    db = Database(**vars(sql_conf))
    run_conf = RunConfig(dataset_id=dataset.id)

    for method, n_parts in METHOD_HYPERPARTS.items():
        run_conf.methods = [method]
        run_id = enter_data(sql_conf, run_conf)

        assert db.get_datarun(run_id)
        with db_session(db):
            run = db.get_datarun(run_id)
            assert run.dataset.id == dataset.id
            assert len(run.hyperpartitions) == n_parts


def test_enter_data_all(dataset):
    sql_conf = SQLConfig(database=DB_PATH)
    db = Database(**vars(sql_conf))
    run_conf = RunConfig(dataset_id=dataset.id,
                         methods=METHOD_HYPERPARTS.keys())

    run_id = enter_data(sql_conf, run_conf)

    with db_session(db):
        run = db.get_datarun(run_id)
        assert run.dataset.id == dataset.id
        assert len(run.hyperpartitions) == sum(METHOD_HYPERPARTS.values())


def test_run_per_partition(dataset):
    sql_conf = SQLConfig(database=DB_PATH)
    db = Database(**vars(sql_conf))
    run_conf = RunConfig(dataset_id=dataset.id, methods=['logreg'])

    run_ids = enter_data(sql_conf, run_conf, run_per_partition=True)

    with db_session(db):
        runs = []
        for run_id in run_ids:
            run = db.get_datarun(run_id)
            if run is not None:
                runs.append(run)

        assert len(runs) == METHOD_HYPERPARTS['logreg']
        assert all([len(run.hyperpartitions) == 1 for run in runs])



#!/usr/bin/python2.7
from __future__ import print_function
import argparse
import os
import yaml
from collections import defaultdict
from os.path import join

from atm.config import *
from atm.database import Database
from atm.enter_data import enter_data
from atm.utilities import download_file_s3
from atm.worker import work

from utilities import *


CONF_DIR = os.path.join(PROJECT_ROOT, 'config/test/')
DATA_DIR = os.path.join(PROJECT_ROOT, 'data/test/')
RUN_CONFIG = join(CONF_DIR, 'run-default.yaml')
SQL_CONFIG = join(CONF_DIR, 'sql-sqlite.yaml')
DATASETS = [
    'iris.data.csv',
    'pollution_1.csv',
]


parser = argparse.ArgumentParser(description='''
Run a single end-to-end test with 10 sample datasets.
The script will create a datarun for each dataset, then run a worker until the
jobs are finished.
''')
parser.add_argument('--processes', help='number of processes to run concurrently',
                    type=int, default=1)
parser.add_argument('--method', help='code for method to test')
parser.add_argument('--method-path', help='path to JSON config for method to test')

args = parser.parse_args()
sql_config, run_config, aws_config, _ = load_config(sql_path=SQL_CONFIG,
                                                    run_path=RUN_CONFIG)
db = Database(**vars(sql_config))

print('creating dataruns...')
datarun_ids = []
for ds in DATASETS:
    run_config.train_path = join(DATA_DIR, ds)
    if args.method:
        run_config.methods = [args.method]
    else:
        run_config.methods = METHODS
    datarun_ids.extend(enter_data(sql_config, run_config, aws_config,
                                  run_per_partition=True))

print('computing on dataruns', datarun_ids)
work_parallel(db=db, datarun_ids=datarun_ids, aws_config=aws_config,
              n_procs=args.processes)

print('workers finished.')

for rid in datarun_ids:
    print_hp_summary(db, rid)



#!/usr/bin/python2.7
from __future__ import print_function

import argparse
import datetime
import os
import warnings

from atm import PROJECT_ROOT
from atm.config import (add_arguments_aws_s3, add_arguments_logging,
                        add_arguments_sql, load_config, initialize_logging)
from atm.database import Database
from atm.worker import Worker, work

warnings.filterwarnings('ignore')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Add more classifiers to database')
    add_arguments_sql(parser)
    add_arguments_aws_s3(parser)
    add_arguments_logging(parser)

    # add worker-specific arguments
    parser.add_argument('--cloud-mode', action='store_true', default=False,
                        help='Whether to run this worker in cloud mode')
    parser.add_argument('--dataruns', help='Only train on dataruns with these ids',
                        nargs='+')
    parser.add_argument('--time', help='Number of seconds to run worker', type=int)
    parser.add_argument('--choose-randomly', action='store_true',
                        help='Choose dataruns to work on randomly (default = sequential order)')
    parser.add_argument('--no-save', dest='save_files', default=True,
                        action='store_const', const=False,
                        help="don't save models and metrics at all")

    # parse arguments and load configuration
    args = parser.parse_args()

    # default logging config is different if initialized from the command line
    if args.log_config is None:
        args.log_config = os.path.join(PROJECT_ROOT,
                                       'config/templates/log-script.yaml')

    sql_config, _, aws_config, log_config = load_config(**vars(args))
    initialize_logging(log_config)

    # let's go
    work(db=Database(**vars(sql_config)),
         datarun_ids=args.dataruns,
         choose_randomly=args.choose_randomly,
         save_files=args.save_files,
         cloud_mode=args.cloud_mode,
         aws_config=aws_config,
         log_config=log_config,
         total_time=args.time,
         wait=False)



"""
Stripped down and modified from the example at:
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='atm',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.9',

    description='a multi-user, multi-data model exploration system',

    # The project's main homepage.
    url='https://github.com/HDI-project/ATM',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        # TODO: python 3 support
    ],

    # What does your project relate to?
    keywords='machine learning hyperparameters tuning classification',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    python_requires='>=2.7',

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'sqlalchemy>=1.1',
        'numpy>=1.13',
        'boto>=2.48',
        'pandas>=0.22',
        'scikit-learn>=0.18',
        'scipy>=0.19',
        'sklearn-pandas>=1.5',
        'mysqlclient>=1.2',
        'pyyaml>=3.12',
        'joblib>=0.11',
        'future>=0.16',
        'btb>=0.0.1',
    ],
    # TODO: this is deprecated. Figure out how to accomplish the same thing with
    # the proper tools.
    # https://www.python.org/dev/peps/pep-0440/#direct-references
    dependency_links=[
        'git+ssh://git@github.com/hdi-project/btb.git#egg=btb-0.0.1',
    ],

    # This variable is used to specify requirements for *this file* to run.
    setup_requires=[],

    test_suite='atm/tests',
    tests_require=[
        'pytest>=3.2',
        'mock>=2',
        'pytest-xdist>=1.20',
        'pytest-runner>=3',
        'pytest-cov>=2.5',
    ]
)

